#include "engine.h"

#include <QDebug>

Engine::Engine(QObject *parent) : QObject(parent)
{
    // Set up actions
    m_zoomOutAction = new QAction("Zoom Out", this);
    m_zoomOutAction->setShortcut(QKeySequence::StandardKey::ZoomOut);
    m_zoomOutAction->setIcon(QIcon(":/icons/zoom_out"));
    connect(m_zoomOutAction, &QAction::triggered, this, &Engine::zoomOut);

    m_centerOnPlayheadAction = new QAction("Center on Playhead", this);
    m_centerOnPlayheadAction->setIcon(QIcon(":/icons/center"));

    m_zoomInAction = new QAction("Zoom In", this);
    m_zoomInAction->setShortcut(QKeySequence::StandardKey::ZoomIn);
    m_zoomInAction->setIcon(QIcon(":/icons/zoom_in"));
    connect(m_zoomInAction, &QAction::triggered, this, &Engine::zoomIn);

    m_stepBackAction = new QAction("Step Back", this);
    m_stepBackAction->setIcon(QIcon(":/icons/chev_left"));
    connect(m_stepBackAction, &QAction::triggered, this, &Engine::stepBack);

    m_playAction = new QAction("Play/Pause", this);
    m_playAction->setIcon(QIcon(":/icons/play"));
    connect(m_playAction, &QAction::triggered, this, &Engine::togglePlayPause);

    m_stepForwardAction = new QAction("Step Forward", this);
    m_stepForwardAction->setIcon(QIcon(":/icons/chev_right"));
    connect(m_stepForwardAction, &QAction::triggered, this, &Engine::stepForward);

    m_plotFile = nullptr;
}

void Engine::clearPlotFile()
{
    // Delete existing plot file object
    if (m_plotFile != nullptr)
        m_plotFile->deleteLater();


    setPlotIndex(0);
    setPlotXQuantityIndex(0);
    setPlotYQuantityIndex(0);
    setPlotSelected(false);
    setPlotLocked(false);
}

void Engine::loadPlotFile(const QString& path)
{
    // Delete existing plot file object
    clearPlotFile();

    m_plotFile = new RmcPlotFile(this);
    connect(m_plotFile, &RmcPlotFile::loadedSuccessfully, this, [=]() {
        if (m_plotFile->plots().count() > 0)
        {
            connect(m_plotFile->plots().at(plotIndex()), &RmcPlot::visibilityChanged, this, [=]() {emit plotVisibilityChanged();});
        }
        emit plotLoaded();

        if (m_emitContentsChanged)
        {
            emit contentsChanged();
        }
    });

    connect(m_plotFile, &RmcPlotFile::errorOccurred, this, [=](auto errorString)
    {
        qDebug() << "Error occurred while loading plot file";
        emit plotLoadingFailed(errorString);
        clearPlotFile();
    });

    m_plotFile->loadPlotFile(path);
}

void Engine::play()
{
    setState(PlayState::Playing);
    m_lastTimerTick = QDateTime::currentMSecsSinceEpoch();
    m_playAction->setIcon(QIcon(":/icons/pause"));

    baseTimelineItem()->mediaPlayer()->play();

    emit videoPlaying();
}

void Engine::pause()
{
    setState(PlayState::Paused);
    m_playAction->setIcon(QIcon(":/icons/play"));

    baseTimelineItem()->mediaPlayer()->pause();
    for (auto i : qAsConst(m_timelineItems))
    {
        i->setPosition(baseTimelineItem()->mediaPlayer()->position());
    }

    emit videoPaused();
}

void Engine::togglePlayPause()
{
    if (state() == PlayState::Paused) play();
    else pause();
}

void Engine::setPlotFile(RmcPlotFile* plotFile)
{
    m_plotFile = plotFile;
    m_plotFile->setParent(this);
    emit plotLoaded();
}

void Engine::setBaseTimelineItem(TimelineItem *item)
{
    m_baseTimelineItem = item;
    m_baseTimelineItem->setParent(this);
    m_baseTimelineItem->setBaseLayer(true);

    connect(m_baseTimelineItem->mediaPlayer(), &QMediaPlayer::positionChanged, this, &Engine::positionChanged);
    connect(m_baseTimelineItem->mediaPlayer(), &QMediaPlayer::durationChanged, this, &Engine::positionChanged);

    connect(item->mediaPlayer(), &QMediaPlayer::mediaStatusChanged,
            this,
        [=](QMediaPlayer::MediaStatus status)
        {
            if (status == QMediaPlayer::MediaStatus::InvalidMedia)
            {
                qDebug() << "INVALID VIDEO FOR BASE ITEM!!!" << item->url().toString();
                // Loading the base item failed, so fall back to the previous one
                emit invalidTimelineItemLoaded(item);
            }

            else
            {
                setLeftVisibleBound(0);
                setRightVisibleBound(durationSecs());
            }
        }
    );

    if (m_emitContentsChanged)
    {
        emit contentsChanged();
    }
}

void Engine::addTimelineItem(TimelineItem *item)
{
    item->setParent(this);
    item->setBaseLayer(false);
    m_timelineItems.append(item);

    connect(item, &TimelineItem::offsetChanged, this,
        [=](double)
        {
            emit timingsChanged();
        }
    );

    connect(item, &TimelineItem::selectedChanged, this,
        [=](bool selected)
        {
            if (selected)
            {
                emit selectionChanged(item);
            }
        }
    );

    connect(item->mediaPlayer(), &QMediaPlayer::mediaStatusChanged, this,
        [=](QMediaPlayer::MediaStatus status)
        {
            if (status == QMediaPlayer::MediaStatus::InvalidMedia)
            {
                emit invalidTimelineItemLoaded(item);
            }
        }
    );

    if (m_emitContentsChanged)
    {
        emit contentsChanged();
    }
}

void Engine::deleteTimelineItem(TimelineItem *item)
{
    m_timelineItems.removeAll(item);
    item->deleteLater();
    emit contentsChanged();
}

void Engine::writeXml(QIODevice* f)
{
    QXmlStreamWriter writer(f);
    writer.setAutoFormatting(true);
    writer.writeStartDocument();

    writer.writeStartElement("VideoSync");

    // Write plot data
    writer.writeStartElement("Plot");
    writer.writeAttribute("Path", plotFile()->url().toString());
    writer.writeAttribute("Index", QString::number(plotIndex()));
    writer.writeAttribute("Offset", QString::number(plotDisplayOffset()));
    writer.writeAttribute("Locked", plotLocked() ? "True" : "False");
    writer.writeAttribute("XQuantity", QString::number(plotXQuantityIndex()));
    writer.writeEndElement();

    // Write base video timeline item
    writer.writeStartElement("BaseVideo");
    writer.writeAttribute("Name", baseTimelineItem()->name());
    writer.writeAttribute("Path", baseTimelineItem()->url().toString());
    writer.writeAttribute("Visible", baseTimelineItem()->visible() ? "True" : "False");
    writer.writeEndElement();

    // Write each video's info
    for (auto& v : timelineItems())
    {
        writer.writeStartElement("Video");
        writer.writeAttribute("Name", v->name());
        writer.writeAttribute("Path", v->url().toString());
        writer.writeAttribute("Offset", QString::number(v->offset()));
        writer.writeAttribute("Locked", v->locked() ? "True" : "False");
        writer.writeAttribute("Visible", v->visible() ? "True" : "False");
        writer.writeEndElement();
    }

    writer.writeEndElement();
    writer.writeEndDocument();
}

void Engine::readXml(QIODevice *f)
{
    QXmlStreamReader reader(f);

    bool offsetOk = false;
    bool plotIndexOk = false;
    bool xQuantIndexOk = false;

    if (m_loadingConfigInfo != nullptr)
    {
        m_loadingConfigInfo->deleteLater();
    }
    m_loadingConfigInfo = new ConfigInfo(this);

    while (!reader.atEnd())
    {
        switch (reader.readNext())
        {
        case QXmlStreamReader::StartElement:
            if (reader.name() == "Plot")
            {
                QString plotPath = reader.attributes().value("Path").toString();
                qreal plotOffset = reader.attributes().value("Offset").toDouble(&offsetOk);
                bool plotLocked = reader.attributes().value("Locked").toString() == "True";
                int plotIndex = reader.attributes().value("Index").toInt(&plotIndexOk);
                int xQuantIndex = reader.attributes().value("XQuantity").toInt(&xQuantIndexOk);

                // Create the plot
                m_loadingConfigInfo->plotInfo()->setPath(plotPath);
                m_loadingConfigInfo->plotInfo()->setOffset(plotOffset);
                m_loadingConfigInfo->plotInfo()->setLocked(plotLocked);
                m_loadingConfigInfo->plotInfo()->setIndex(plotIndex);
                m_loadingConfigInfo->plotInfo()->setXQuantityIndex(xQuantIndex);
            }
            else if (reader.name() == "BaseVideo")
            {
                auto path = reader.attributes().value("Path").toString();
                auto name = reader.attributes().value("Name").toString();
                auto visible = reader.attributes().value("Visible").toString() == "True";

                if (path.isEmpty())
                    break;

                if (name.isEmpty())
                    name = "Base Video";

                // Create the base video
                m_loadingConfigInfo->baseVideoInfo()->setPath(path);
                m_loadingConfigInfo->baseVideoInfo()->setName(name);
                m_loadingConfigInfo->baseVideoInfo()->setVisible(visible);
            }
            else if (reader.name() == "Video")
            {
                auto name = reader.attributes().value("Name").toString();
                auto path = reader.attributes().value("Path").toString();
                auto offset = reader.attributes().value("Offset").toDouble(&offsetOk);
                auto locked = reader.attributes().value("Locked").toString() == "True";
                auto visible = reader.attributes().value("Visible").toString() == "True";

                AdditionalVideoInfo* a = new AdditionalVideoInfo(m_loadingConfigInfo);
                a->setName(name);
                a->setPath(path);
                a->setOffset(offset);
                a->setLocked(locked);
                a->setVisible(visible);
                m_loadingConfigInfo->addAdditionalVideoInfo(a);
            }
        default:
            break;
        }
    }

    qDebug() << "Done reading XML";
    qDebug() << "Plot info:" << m_loadingConfigInfo->plotInfo()->info();

    qDebug() << "Base video info:" << m_loadingConfigInfo->baseVideoInfo()->info();
    qDebug() << "Additional video info:" << m_loadingConfigInfo->additionalVideoInfo().count();
    for (auto& i : m_loadingConfigInfo->additionalVideoInfo())
    {
        qDebug() << i->info();
    }

    m_loadingConfigInfo->dumpObjectTree();


    // Before we start trying to actually load anything, let's make sure
    // that this data is valid.

    // First, we need to have a base video.
    if (m_loadingConfigInfo->baseVideoInfo()->path().isEmpty())
    {
        // No base video, abort!
        qWarning() << "No base video was found! Aborting load...";
        m_loadingConfigInfo->deleteLater();
        return;
    }

    // Second, we need to have a plot.
    if (m_loadingConfigInfo->plotInfo()->path().isEmpty())
    {
        // No plot, abort!
        qWarning() << "No plot was found! Aborting load...";
        m_loadingConfigInfo->deleteLater();
        return;
    }

    loadConfig_baseTimelineItem();
}

void Engine::loadConfig_baseTimelineItem()
{
    // Okay, there was a base video and a plot specified in this file.
    // Next, we need to try to load the base video.
    m_loadingConfigInfo->baseVideo()->setBaseLayer(true);
    m_loadingConfigInfo->baseVideo()->setName(m_loadingConfigInfo->baseVideoInfo()->name());
    m_loadingConfigInfo->baseVideo()->setVisible(m_loadingConfigInfo->baseVideoInfo()->visible());

    // Using strategy from https://forum.qt.io/topic/67272/how-to-create-a-single-shot-one-time-connection-to-a-lambda
    m_loadBaseVideoConnection = connect(m_loadingConfigInfo->baseVideo()->mediaPlayer(), &QMediaPlayer::mediaStatusChanged, this, [=](auto status)
    {
        // Media was invalid.
        // We can't load the base video file, so we need to abort the
        // whole loading process.
        if (status == QMediaPlayer::MediaStatus::InvalidMedia)
        {
            qWarning() << "Base video could not be loaded! Aborting load...";
            loadConfig_unload();
            QObject::disconnect(m_loadBaseVideoConnection);
        }

        // Media was buffered!
        // We can now move on to the next step: loading the plot file.
        else if (status == QMediaPlayer::MediaStatus::BufferedMedia)
        {
            qDebug() << "Base video was successfully loaded!";
            QObject::disconnect(m_loadBaseVideoConnection);
            loadConfig_plot();
        }
    });
    m_loadingConfigInfo->baseVideo()->loadVideo(m_loadingConfigInfo->baseVideoInfo()->path());
}

void Engine::loadConfig_plot()
{
    qDebug() << "Now attempting to load plot...";

    // Start trying to load the plot file.
    m_loadPlotConnection = connect(m_loadingConfigInfo->plotFile(), &RmcPlotFile::loadedSuccessfully, this, [=]()
    {
        qDebug() << "Plot was successfully loaded!";
        loadConfig_addlVideos();
    });

    m_loadPlotFailedConnection = connect(m_loadingConfigInfo->plotFile(), &RmcPlotFile::errorOccurred, this, [=](auto errorString)
    {
        emit plotLoadingFailed(errorString);
        loadConfig_unload();
    });

    m_loadingConfigInfo->plotFile()->loadPlotFile(m_loadingConfigInfo->plotInfo()->path());
}

void Engine::loadConfig_addlVideos()
{
    *m_loadingVideosLoaded = 0;
    int totalVideosToLoad = m_loadingConfigInfo->additionalVideoInfo().count();

    for (auto& singleVideoInfo : m_loadingConfigInfo->additionalVideoInfo())
    {
        auto singleVideo = new TimelineItem(m_loadingConfigInfo);
        singleVideo->setBaseLayer(false);
        singleVideo->setName(singleVideoInfo->name());
        singleVideo->setOffset(singleVideoInfo->offset());
        singleVideo->setVisible(singleVideoInfo->visible());
        singleVideo->setLocked(singleVideoInfo->locked());
        m_loadingConfigInfo->addAdditionalVideo(singleVideo);

        QMetaObject::Connection* const c3 = new QMetaObject::Connection;
        *c3 = connect(singleVideo->mediaPlayer(), &QMediaPlayer::mediaStatusChanged, this, [=](auto status)
        {
            if (status == QMediaPlayer::MediaStatus::InvalidMedia)
            {
                qWarning() << "Video with name" << singleVideoInfo->name() << "could not be loaded!";
                // For now, if a video can't be found, let's abort the whole loading
                // operation. In the future, it might be nice to preserve what we can
                // load, and just indicate to the user that a particular video was
                // unable to load.
                loadConfig_unload();
            }
            else if (status == QMediaPlayer::MediaStatus::BufferedMedia)
            {
                (*m_loadingVideosLoaded)++;
                qDebug() << "Video with name" << singleVideoInfo->name() << "loaded, that's" << *m_loadingVideosLoaded << "out of" << totalVideosToLoad;
                if (*m_loadingVideosLoaded >= totalVideosToLoad)
                {
                    qDebug() << "All videos have been loaded!";
                    loadConfig_putEverythingInEngine();
                }
            }
        });
        m_addlVideoConnections.append(c3);
        singleVideo->loadVideo(singleVideoInfo->path());
    }
}

void Engine::loadConfig_putEverythingInEngine()
{
    m_emitContentsChanged = false;
    qDebug() << "Now putting everything into the engine itself...";
    // First, set the base timeline item
    setBaseTimelineItem(m_loadingConfigInfo->baseVideo());
    qDebug() << "Base timeline item is in engine!";

    // Second, set the plot item
    setPlotFile(m_loadingConfigInfo->plotFile());
    qDebug() << "Plot file is in engine!";

    // Third, add all of the timeline items
    for (auto& item : m_timelineItems)
    {
        item->deleteLater();
    }
    m_timelineItems.clear();

    for (auto& item : m_loadingConfigInfo->additionalVideos())
    {
        addTimelineItem(item);
    }
    qDebug() << "All additional timeline items are in engine!";
    qDebug() << "Everything should now be loaded! :)";

    m_emitContentsChanged = true;
    emit contentsChanged();

    loadConfig_unload();
}

void Engine::loadConfig_unload()
{
    for (auto i : qAsConst(m_addlVideoConnections))
    {
        QObject::disconnect(*i);
        delete i;
    }
    m_addlVideoConnections.clear();

    m_loadingConfigInfo->deleteLater();
    m_loadingConfigInfo = nullptr;
}
