#ifndef ENGINE_H
#define ENGINE_H

#include <QObject>
#include <QMediaPlayer>
#include <QAction>
#include <QTimer>
#include <QDateTime>
#include <QXmlStreamWriter>

#include "Models/Timeline/timelineitem.h"
#include "Utilities/graphicsmath.h"
#include "Utilities/PlotParsing/rmcplotfile.h"

class PlotInfo : public QObject
{
    Q_OBJECT
public:
    PlotInfo(QObject* parent = nullptr) : QObject(parent) {}

    const QString& path() { return m_path; }
    void setPath(const QString& path) { m_path = path; }

    qreal offset() { return m_offset; }
    void setOffset(qreal offset) { m_offset = offset; }

    bool locked() { return m_locked; }
    void setLocked(bool locked) { m_locked = locked; }

    int index() { return m_index; }
    void setIndex(int index) { m_index = index; }

    int xQuantityIndex() { return m_xQuantIndex; }
    void setXQuantityIndex(int index) { m_xQuantIndex = index; }

    QString info()
    {
        return QString("PlotInfo(path=%1, offset=%2, locked=%3)").arg(
                    m_path,
                    QString::number(m_offset),
                    m_locked ? "true" : "false"
                    );
    }

private:
    QString m_path;
    qreal m_offset;
    bool m_locked;
    int m_index;
    int m_xQuantIndex;
};

class BaseVideoInfo : public QObject
{
    Q_OBJECT
public:
    BaseVideoInfo(QObject* parent = nullptr) : QObject(parent) {}

    QString path() { return m_path; }
    void setPath(const QString& path) { m_path = path; }

    QString name() { return m_name; }
    void setName(const QString& name) { m_name = name; }

    bool visible() { return m_visible; }
    void setVisible(bool visible) { m_visible = visible; }

    QString info()
    {
        return QString("BaseVideoInfo(path=%1, name=%2, visible=%3)").arg(
                    m_path,
                    m_name,
                    m_visible ? "true" : "false"
                    );
    }

private:
    QString m_path;
    QString m_name;
    bool m_visible;
};

class AdditionalVideoInfo : public QObject
{
    Q_OBJECT
public:
    AdditionalVideoInfo(QObject* parent = nullptr) : QObject(parent) {}

    QString path() { return m_path; }
    void setPath(const QString& path) { m_path = path; }

    QString name() { return m_name; }
    void setName(const QString& name) { m_name = name; }

    qreal offset() { return m_offset; }
    void setOffset(qreal offset) { m_offset = offset; }

    bool locked() { return m_locked; }
    void setLocked(bool locked) { m_locked = locked; }

    bool visible() { return m_visible; }
    void setVisible(bool visible) { m_visible = visible; }

    QString info()
    {
        return QString("AdditionalVideoInfo(path=%1, name=%2, offset=%3, locked=%4, visible=%5)").arg(
                    m_path,
                    m_name,
                    QString::number(m_offset),
                    m_locked ? "true" : "false",
                    m_visible ? "true" : "false"
                    );
    }

private:
    QString m_path;
    QString m_name;
    qreal m_offset;
    bool m_locked;
    bool m_visible;
};

class ConfigInfo : public QObject
{
    Q_OBJECT
public:
    ConfigInfo(QObject* parent = nullptr) : QObject(parent),
        m_baseVideoInfo(new BaseVideoInfo(this)), m_plotInfo(new PlotInfo(this)),
        m_baseVideo(new TimelineItem(this)), m_plot(new RmcPlotFile(this)) {}

    ~ConfigInfo()
    {
        qDebug() << "Destroying config info";
    }

    BaseVideoInfo* baseVideoInfo() { return m_baseVideoInfo; }
    PlotInfo* plotInfo() { return m_plotInfo; }
    const QList<AdditionalVideoInfo*>& additionalVideoInfo() { return m_addlVideoInfo; }

    void addAdditionalVideoInfo(AdditionalVideoInfo* v) { m_addlVideoInfo.append(v); }
    void addAdditionalVideo(TimelineItem* v) { m_addlVideos.append(v); }

    void removeAdditionalVideo(TimelineItem* v) { m_addlVideos.removeAll(v); }
    TimelineItem* baseVideo() { return m_baseVideo; }
    RmcPlotFile* plotFile() { return m_plot; }
    const QList<TimelineItem*>& additionalVideos() { return m_addlVideos; }

private:
    BaseVideoInfo* m_baseVideoInfo = nullptr;
    PlotInfo* m_plotInfo = nullptr;
    QList<AdditionalVideoInfo*> m_addlVideoInfo;

    TimelineItem* m_baseVideo = nullptr;
    RmcPlotFile* m_plot = nullptr;
    QList<TimelineItem*> m_addlVideos;
};

class Engine : public QObject
{
    Q_OBJECT
public:
    enum PlayState {
        Paused = 0,
        Playing = 1
    };

    enum WorkspaceMode
    {
        Synchronize = 0,
        Analyze = 1
    };

    explicit Engine(QObject *parent = nullptr);

    QAction& zoomOutAction() { return *m_zoomOutAction; }
    QAction& centerOnPlayheadAction() { return *m_centerOnPlayheadAction; }
    QAction& zoomInAction() { return *m_zoomInAction; }

    QAction& stepBackAction() { return *m_stepBackAction; }
    QAction& playAction() { return *m_playAction; }
    QAction& stepForwardAction() { return *m_stepForwardAction; }

    qreal currentTime()
    {
        if (baseTimelineItem() == nullptr)
            return 0.0;

        auto p = baseTimelineItem()->mediaPlayer();
        return (qreal)p->position() / 1000.0;
    }

    void incrementCurrentTime(qreal amount)
    {
        setCurrentTime(currentTime() + amount);
    }

    qreal leftVisibleBound()
    {
        return m_leftVisBound;
    }
    void setLeftVisibleBound(qreal bound)
    {
        m_leftVisBound = qBound(0., bound, rightVisibleBound() - 0.5);
        emit zoomBoundsChanged();
    }

    qreal rightVisibleBound()
    {
        return qBound(leftVisibleBound() + 0.5, m_rightVisBound, durationSecs());
    }
    void setRightVisibleBound(qreal bound)
    {
        m_rightVisBound = qBound(leftVisibleBound() + 0.5, bound, durationSecs());
        emit zoomBoundsChanged();
    }

    TimelineItem* baseTimelineItem() { return m_baseTimelineItem; }
    const QList<TimelineItem*>& timelineItems() { return m_timelineItems; }

    qreal durationSecs()
    {
        if (baseTimelineItem() == nullptr) return 0.0;
        return baseTimelineItem()->duration();
    }

    TimelineItem* timelineItem(int index)
    {
        return index == 0 ? baseTimelineItem() : timelineItems().at(index - 1);
    }

    void play();
    void pause();

    void setBaseTimelineItem(TimelineItem *item);
    void addTimelineItem(TimelineItem* item);
    void deleteTimelineItem(TimelineItem *item);

    PlayState state() { return m_state; }
    void setState(PlayState state) { m_state = state; }

    WorkspaceMode workspaceMode() { return m_workspaceMode; }
    void setWorkspaceMode(WorkspaceMode mode) { m_workspaceMode = mode; }

    RmcPlotFile* plotFile() { return m_plotFile; }
    void setPlotFile(RmcPlotFile* plotFile);

    int plotIndex() { return m_plotIndex; }
    void setPlotIndex(int index)
    {
        m_plotIndex = index;
        emit plotIndexChanged();
    }

    int plotXQuantityIndex() { return m_plotXQuantityIndex; }
    void setPlotXQuantityIndex(int index)
    {
        m_plotXQuantityIndex = index;
        emit plotXQuantityIndexChanged(m_plotXQuantityIndex);
    }

    RmcPlotQuantity *plotXQuantity()
    {
        if (currentPlot() == nullptr)
            return nullptr;

        return currentPlot()->quantities().at(plotXQuantityIndex());
    }
    void setPlotXQuantity(RmcPlotQuantity *quantity)
    {
        auto index = currentPlot()->quantities().indexOf(quantity);
        setPlotXQuantityIndex(index);
    }

    int plotYQuantityIndex() { return m_plotYQuantityIndex; }
    void setPlotYQuantityIndex(int index)
    {
        m_plotYQuantityIndex = index;
        emit plotYQuantityIndexChanged();
    }

    RmcPlotQuantity *plotYQuantity()
    {
        if (currentPlot() == nullptr)
            return nullptr;

        if (currentPlot()->quantities().count() == 0)
            return nullptr;

        return currentPlot()->quantities().value(plotYQuantityIndex());
    }
    void setPlotYQuantity(RmcPlotQuantity *quantity)
    {
        auto index = currentPlot()->quantities().indexOf(quantity);
        setPlotYQuantityIndex(index);
    }

    const QVector<RmcPlotQuantity*> quantitiesForCurrentScaleGroup()
    {
        QVector<RmcPlotQuantity*> v;

        // plotYQuantity() will return a nullptr if there's no plot
        // or quantities loaded. In that case, we don't have any
        // quantities to give, so we'll just return an empty vector.
        auto yQuantity = plotYQuantity();
        if (yQuantity == nullptr)
            return v;

        auto scaleGroup = plotYQuantity()->scaleGroup();

        for (auto q : currentPlot()->quantities())
        {
            if (q->scaleGroup() == scaleGroup) v.append(q);
        }
        return v;
    }

    const QVector<RmcPlotQuantity*> quantitiesForCurrentXAxisScaleGroup()
    {
        QVector<RmcPlotQuantity*> v;
        auto scaleGroup = plotXQuantity()->scaleGroup();

        for (auto q : currentPlot()->quantities())
        {
            if (q->scaleGroup() == scaleGroup) v.append(q);
        }
        return v;
    }

    bool plotSelected() { return m_plotSelected; }

    bool plotLocked() { return m_plotLocked; }

    qreal plotOffset() { return m_plotOffset; }
    qreal plotDisplayOffset() { return m_plotOffset + m_plotTempTranslation; }
    void setPlotOffset(qreal offset)
    {
        m_plotOffset = offset;
        emit timingsChanged();
    }

    qreal plotTemporaryTranslation() { return m_plotTempTranslation; }
    void setPlotTemporaryTranslation(qreal translation) { m_plotTempTranslation = translation; }
    void commitPlotTemporaryTranslation()
    {
        setPlotOffset(m_plotOffset + m_plotTempTranslation);
        setPlotTemporaryTranslation(0);
    }
    void revertPlotTemporaryTranslation() { setPlotTemporaryTranslation(0); }


    QImage* getTYPlot(QSize size)
    {
        return plotFile()->getTYImage(size.width(), size.height(), plotIndex());
    }

    QImage* getXYPlot(QSize size)
    {
        return plotFile()->getXYImage(size.width(), size.height(), plotIndex(), plotXQuantityIndex());
    }

    qreal currentPlotDuration()
    {
        return currentPlot()->duration();
    }

    RmcPlot* currentPlot()
    {
        if (plotFile() == nullptr || plotFile()->plots().count() == 0)
            return nullptr;

        return plotFile()->plots().at(plotIndex());
    }

    qreal scrollBarRange()
    {
        return leftVisibleBound() + rightVisibleBound();
    }
    qreal centerOfScrollBar()
    {
        return scrollBarRange() / 2;
    }
    qreal zoomStep()
    {
        return scrollBarRange() * m_zoomStepMultiplier;
    }

    void writeXml(QIODevice* f);
    void readXml(QIODevice* f);

    void clearPlotFile();
    void loadPlotFile(const QString& path);

private:
    void loadConfig_baseTimelineItem();
    void loadConfig_plot();
    void loadConfig_addlVideos();
    void loadConfig_putEverythingInEngine();
    void loadConfig_unload();

    // When we're loading in a bunch of stuff at the same time,
    // we don't want the contentsChanged signal to be emitted over
    // and over again - only after the last change is made.
    // To do this, we'll check the value of this boolean before
    // emitting the signal.
    bool m_emitContentsChanged = true;

protected:
    QAction* m_zoomOutAction = nullptr;
    QAction* m_centerOnPlayheadAction = nullptr;
    QAction* m_zoomInAction = nullptr;

    QAction* m_stepBackAction = nullptr;
    QAction* m_playAction = nullptr;
    QAction* m_stepForwardAction = nullptr;

    qreal m_leftVisBound = 0.0;
    qreal m_rightVisBound = 8.704;

    const qreal m_zoomStepMultiplier = 0.05;

    TimelineItem* m_savedBaseTimelineItem = nullptr;
    TimelineItem* m_baseTimelineItem = nullptr;
    QList<TimelineItem*> m_timelineItems;

    qint64 m_lastTimerTick;

    bool m_plotSelected = false;
    bool m_plotLocked = false;
    qreal m_plotOffset = 0.0;
    qreal m_plotTempTranslation = 0.0;

    PlayState m_state = PlayState::Paused;
    WorkspaceMode m_workspaceMode = WorkspaceMode::Synchronize;

    RmcPlotFile* m_plotFile = nullptr;
    int m_plotIndex = 0;
    int m_plotXQuantityIndex = 0;
    int m_plotYQuantityIndex = 0;

    ConfigInfo* m_loadingConfigInfo = nullptr;
    QMetaObject::Connection m_loadBaseVideoConnection;
    QMetaObject::Connection m_loadPlotConnection;
    QMetaObject::Connection m_loadPlotFailedConnection;
    QVector<QMetaObject::Connection*> m_addlVideoConnections;

    int* m_loadingVideosLoaded = new int(0);

public:
    static Engine* engine;

protected slots:
    void togglePlayPause();

    void stepBack()
    {
        incrementCurrentTime(-0.001f);
    }

    void stepForward()
    {
        incrementCurrentTime(0.001f);
    }

    void zoomIn()
    {
        setLeftVisibleBound(leftVisibleBound() + zoomStep());
        setRightVisibleBound(rightVisibleBound() - zoomStep());

        emit zoomBoundsChanged();
    }

    void zoomOut()
    {
        setLeftVisibleBound(leftVisibleBound() - zoomStep());
        setRightVisibleBound(rightVisibleBound() + zoomStep());

        emit zoomBoundsChanged();
    }

public slots:
    void setCurrentTime(qreal time)
    {
        time = qBound(0., time, durationSecs());
        auto intTime = (qint64)(time * 1000);
        baseTimelineItem()->setPosition(intTime);
        for (auto& i : m_timelineItems) { i->setPosition(intTime); }
    }

    void setPlotSelected(bool selected)
    {
        m_plotSelected = selected;
        emit selectionChanged(nullptr);
    }

    void setPlotLocked(bool locked) { m_plotLocked = locked; }

signals:
    void positionChanged();
    void contentsChanged();
    void timingsChanged();
    void selectionChanged(TimelineItem* selectedItem);

    void zoomBoundsChanged();

    void plotOffsetChanged();
    void plotLoaded();
    void plotVisibilityChanged();
    void plotIndexChanged();
    void plotXQuantityIndexChanged(int);
    void plotYQuantityIndexChanged();

    void videoPlaying();
    void videoPaused();

    void invalidTimelineItemLoaded(TimelineItem* item);
    void plotLoadingFailed(QString errorString);
};
#endif // ENGINE_H
