#include "rmcplotfile.h"

const QString RmcPlotFile::filter_str = QString("Filter");
const QString RmcPlotFile::graph_str = QString("Graph");
const QString RmcPlotFile::autoScale_str = QString("AutoScale");
const QString RmcPlotFile::centerOn0_str = QString("CenterOn0");
const QString RmcPlotFile::minRange_str = QString("MinRange");
const QString RmcPlotFile::bottomScale_str = QString("BottomScale");
const QString RmcPlotFile::topScale_str = QString("TopScale");
const QString RmcPlotFile::autoGroup_str = QString("AutoGroup");
const QString RmcPlotFile::scaleGroup_str = QString("ScaleGroup");
const QString RmcPlotFile::railToRail_str = QString("RailToRail");

RmcPlotFile::RmcPlotFile(QObject *parent) : QThread(parent)
{
    connect(this, &QThread::finished, this, &RmcPlotFile::onFinished);
}

void RmcPlotFile::onFinished()
{
    if (!m_errorString.isEmpty())
    {
        qCritical() << "Error loading plot file!" << m_errorString;
        for (auto i : qAsConst(m_plots))
        {
            i->deleteLater();
        }
        emit errorOccurred(m_errorString);
    }
    else
    {
        for (auto i : qAsConst(m_plots))
        {
            i->setParent(this);
        }
        emit loadedSuccessfully();
    }
}

void RmcPlotFile::loadPlotFile(const QString& path)
{
    m_url.setUrl(path);
    start();
}

void RmcPlotFile::run()
{
    QFile file(m_url.toString());
    if (!file.open(QIODevice::ReadOnly))
    {
        qDebug() << "error opening file!" << file.errorString() << file.error();
        m_errorString = file.errorString();
        return;
    }

    // Get the file info to determine whether this is an
    // RMCPLOTS (uncompressed) or RMCPLOTX (compressed) file.
    QFileInfo fileInfo(file);
    QString suffix = fileInfo.completeSuffix();
    if (suffix == "rmcplots")
    {
        qDebug() << "it's an RMCPLOTS file";
        parseUncompressedPlot(file);
    }
    else if (suffix == "rmcplotx")
    {
        qDebug() << "it's an RMCPLOTX file - need to decompress it!";
        parseCompressedPlot(file);
    }

    file.close();
}

void RmcPlotFile::parseUncompressedPlot(QFile& file)
{
    QXmlStreamReader reader(&file);
    while (!reader.atEnd())
    {
        reader.readNext();
        parseNextElement(&reader);
    }

}

void RmcPlotFile::parseCompressedPlot(QFile& file)
{
    QXmlStreamReader reader;

    // Set up zlib stuff
    z_stream zStream;
    zStream.zalloc = Z_NULL;
    zStream.zfree = Z_NULL;
    zStream.opaque = Z_NULL;
    zStream.avail_in = 0;
    zStream.next_in = Z_NULL;
    int zRet = inflateInit2(&zStream, 16 + MAX_WBITS);

    if (zRet == Z_MEM_ERROR)
    {
        m_errorString = QString("Not enough memory to decompress plot file (inflateInit2() returned Z_MEM_ERROR)");
        return;
    }
    else if (zRet == Z_VERSION_ERROR)
    {
        m_errorString = QString("Current version of zlib is incompatible (inflateInit2() returned Z_VERSION_ERROR)");
        return;
    }
    else if (zRet == Z_STREAM_ERROR)
    {
        m_errorString = QString("Input parameters for decompression are invalid (inflateInit2() returned Z_STREAM_ERROR)");
        return;
    }

    unsigned long compressedAmountRead;
    const unsigned long long amountOfCompressedDataToRead = 1024;
    unsigned char* compressedData = nullptr;

    while (!reader.atEnd())
    {
        auto next = reader.readNext();

        // Case of buffer not being full enough
        // We need to give the buffer more stuff to read
        if (next == QXmlStreamReader::Invalid && reader.error() == QXmlStreamReader::PrematureEndOfDocumentError)
        {
            zStream.avail_out = amountOfCompressedDataToRead;
            unsigned char pachData[amountOfCompressedDataToRead];
            zStream.next_out = pachData;

            if (zStream.avail_in == 0)
            {
                if (compressedData != nullptr) delete[] compressedData;
                compressedData = new unsigned char[amountOfCompressedDataToRead];
                zStream.next_in = compressedData;

                compressedAmountRead = file.read(reinterpret_cast<char*>(compressedData), amountOfCompressedDataToRead);
                emit progressLoading(file.pos(), file.size());

                zStream.avail_in = compressedAmountRead;
            }

            // Do the decompression
            int result = inflate(&zStream, Z_NO_FLUSH);

            if (result == Z_STREAM_END)
            {
                if (zStream.avail_out < amountOfCompressedDataToRead)
                {
                    inflateEnd(&zStream);
                }
            }
            else if (result == Z_DATA_ERROR || result == Z_STREAM_ERROR)
            {
                inflateEnd(&zStream);

                switch (result)
                {
                case Z_DATA_ERROR:
                    m_errorString = QString("Compressed data was corrupted, details: \"%1\" (inflate() returned Z_DATA_ERROR)").arg(zStream.msg);
                    break;
                case Z_STREAM_ERROR:
                    m_errorString = QString("Stream structure was inconsistent (inflate() returned Z_STREAM_ERROR)");
                    break;
                default:
                    break;
                }
            }

            auto data = QByteArray(reinterpret_cast<char*>(pachData), amountOfCompressedDataToRead - zStream.avail_out);
            reader.addData(data);
            continue;
        }
        parseNextElement(&reader);

        if (reader.hasError())
        {
            m_errorString = reader.errorString();
            break;
        }

    }
    qDebug() << "reached end of reader";
}

void RmcPlotFile::parseNextElement(QXmlStreamReader* reader)
{
    switch (reader->tokenType())
    {
    case QXmlStreamReader::NoToken: break;
    case QXmlStreamReader::Invalid:
        // If we got here, then there was a legitimate error
        // reading the XML file, and we need to abort.
        qDebug() << "Error loading" << m_url.toString() << " on line" << reader->lineNumber() << "col" << reader->columnNumber() << ":"
                 << reader->errorString() << "(error code" << reader->error() << ")";
        break;
    case QXmlStreamReader::StartDocument: break;
    case QXmlStreamReader::EndDocument: break;
    case QXmlStreamReader::StartElement:
        if (reader->name() == "PlotFolderRoot")
        {
            m_name = reader->attributes().value("Name").toString();
        }
        else if (reader->name() == "Plot") handlePlotElement(reader);
        else if (reader->name() == "Controller") handleControllerElement(reader);
        else if (reader->name() == "Item") handleItemElement(reader);
        else if (reader->name() == "Quantity") handleQuantityElement(reader);
        else if (reader->name() == "Sample") handleSampleElement(reader);
        break;
    case QXmlStreamReader::EndElement: break;
    default: break;
    }
}

void RmcPlotFile::handlePlotElement(QXmlStreamReader *reader)
{
    auto plot = new RmcPlot;
    m_plots.append(plot);
    m_plots.last()->setName(reader->attributes().value("Name").toString());
    m_plots.last()->setSamplePeriod(reader->attributes().value("SamplePeriodS").toDouble());
}

void RmcPlotFile::handleControllerElement(QXmlStreamReader *reader)
{
    if (reader->attributes().hasAttribute("ControllerType"))
    {
        QString type = reader->attributes().value("ControllerType").toString();
        RmcPlot::ControllerType toSet = RmcPlot::ControllerType::RMC70;
        if (type == "RMC150") toSet = RmcPlot::ControllerType::RMC150;
        else if (type == "RMC200") toSet = RmcPlot::ControllerType::RMC200;

        m_plots.last()->setControllerType(toSet);
    }
    else
    {
        m_plots.last()->setControllerType(RmcPlot::ControllerType::RMC70);
    }
}

void RmcPlotFile::handleItemElement(QXmlStreamReader *reader)
{
    auto newItem = new RmcPlotItem(m_plots.last());
    newItem->setAddress(reader->attributes().value("Address").toString());
    m_plots.last()->appendItem(newItem);
}

void RmcPlotFile::handleQuantityElement(QXmlStreamReader *reader)
{
    QString name, units, formula, type, penString;
    QColor penColor(Qt::GlobalColor::black);
    qreal minRange, bottomScale, topScale;
    int scaleGroupID;
    bool filter, graph, autoScale, centerOn0, railToRail, autoGroup;
    auto a = reader->attributes();


    name = a.value("Name").toString();
    units = a.value("Units").toString();
    formula = a.value("Formula").toString();
    /*bool filter_read = */parseBool(&filter_str, &filter, false, &a);
    penString = a.value("Pen").toString();
    type = a.value("Type").toString();

    /*bool graph_read = */parseBool(&graph_str, &graph, false, &a);
    bool autoScale_read = parseBool(&autoScale_str, &autoScale, true, &a);
    bool centerOn0_read = parseBool(&centerOn0_str, &centerOn0, false, &a);
    bool minRange_read = parseReal(&minRange_str, &minRange, 1.0, &a);
    bool bottomScale_read = parseReal(&bottomScale_str, &bottomScale, FLT_MAX, &a);
    bool topScale_read = parseReal(&topScale_str, &topScale, FLT_MIN, &a);
    bool autoGroup_read = parseBool(&autoGroup_str, &autoGroup, true, &a);
    /*bool scaleGroup_read = */parseInt(&scaleGroup_str, &scaleGroupID, -1, &a);
    bool railToRail_read = parseBool(&railToRail_str, &railToRail, false, &a);

    // Parse the color
    bool colorParsed = false;
    int intColor = penString.remove(0, 3).toInt(&colorParsed, 16);

    if (colorParsed)
    {
        int r = (0x0000FF & intColor) >> 0;
        int g = (0x00FF00 & intColor) >> 8;
        int b = (0xFF0000 & intColor) >> 16;
        penColor = QColor(r, g, b);
    }


    if (autoGroup && !autoGroup_read)
    {
        // Determine quantity type
        auto quantityType = quantityTypeForQuantity(&name, &formula);

        // Based on quantity type, assign scale group
        scaleGroupID = (int)quantityType;

        bool setMinRange = !minRange_read && autoScale && autoScale_read;
        switch (quantityType)
        {
        case RmcPlotFile::QuantityType::Velocity:
        case RmcPlotFile::QuantityType::Acceleration:
        case RmcPlotFile::QuantityType::PressureForceRate:
        case RmcPlotFile::QuantityType::Jerk:
            if (!centerOn0_read) centerOn0 = true;
            if (!railToRail_read) railToRail = false;
            if (setMinRange) minRange = 1.0;
            break;
        case RmcPlotFile::QuantityType::Gains:
            if (!centerOn0_read) centerOn0 = true;
            if (!railToRail_read) railToRail = true;
            if (setMinRange) minRange = 200.0;
            break;
        case RmcPlotFile::QuantityType::ControlOutput75_150:
            if (!centerOn0_read) centerOn0 = true;
            if (!railToRail_read) railToRail = true;
            if (setMinRange) minRange = 20.0;
            break;
        case RmcPlotFile::QuantityType::Position:
        case RmcPlotFile::QuantityType::Counts:
        case RmcPlotFile::QuantityType::RawCounts:
        case RmcPlotFile::QuantityType::PressureForce:
        case RmcPlotFile::QuantityType::SecCounts:
        case RmcPlotFile::QuantityType::SecRawCounts:
        default:
            if (!centerOn0_read) centerOn0 = false;
            if (!railToRail_read) railToRail = false;
            if (setMinRange) minRange = 1.0;
            break;
        }
    }

    scaleGroupID = scaleGroupID == -1 ? m_autogeneratedScaleGroup++ : scaleGroupID;

    auto plot = m_plots.last();

    if (!plot->scaleGroups().contains(scaleGroupID))
    {
        auto newScaleGroup = new RmcScaleGroup;
        newScaleGroup->setAutoScale(autoScale);
        newScaleGroup->setCenterOn0(centerOn0);
        newScaleGroup->setRailToRail(railToRail);
        newScaleGroup->setScales(bottomScale, topScale);
        newScaleGroup->setBottomScaleExplicit(bottomScale_read);
        newScaleGroup->setTopScaleExplicit(topScale_read);
        newScaleGroup->setMinRange(minRange);

        plot->appendScaleGroup(scaleGroupID, newScaleGroup);
    }
    else
    {
        auto existingScaleGroup = plot->scaleGroup(scaleGroupID);
        if (centerOn0_read && existingScaleGroup->centerOn0() != centerOn0)
            existingScaleGroup->setCenterOn0(false);

        if (minRange_read && existingScaleGroup->minRange() != minRange)
            existingScaleGroup->setMinRange(1.0);

        if (railToRail_read && existingScaleGroup->railToRail() != railToRail)
            existingScaleGroup->setRailToRail(false);
    }

    auto quantityObject = new RmcPlotQuantity;
    quantityObject->setName(name);
    quantityObject->setFormula(formula);
    quantityObject->setColor(penColor);
    quantityObject->setType(type);
    quantityObject->setGraphQuantity(graph);
    quantityObject->setAutoScale(autoScale);
    quantityObject->setCenterOn0(centerOn0);
    quantityObject->setMinRange(minRange);
    quantityObject->setBottomScale(bottomScale);
    quantityObject->setTopScale(topScale);
    quantityObject->setAutoGroup(autoGroup);
    quantityObject->setScaleGroup(scaleGroupID);

    connect(quantityObject, &RmcPlotQuantity::visibilityChanged, [=]() {emit m_plots.last()->visibilityChanged();});
    m_plots.last()->appendQuantity(quantityObject);
}

void RmcPlotFile::handleSampleElement(QXmlStreamReader *reader)
{
    int interval = reader->attributes().value("Int").toInt();

    auto newSample = new RmcPlotSample;
    newSample->setInterval(interval);
    m_plots.last()->appendSample(interval, newSample);

    for (int i = 0; i < m_plots.last()->items().size(); i++)
    {
        auto item = m_plots.last()->items().at(i);
        auto stringValue = reader->attributes().value(QString("Item%1").arg(i));

        qreal value;

        // Check prefix - if it's 16# or 0x, then parse it as hex.
        if (stringValue.startsWith("16#"))
        {
            value = static_cast<qreal>(stringValue.mid(3).toUInt(nullptr, 16));
        }
        else if (stringValue.startsWith("0x"))
        {
            value = static_cast<qreal>(stringValue.mid(2).toUInt(nullptr, 16));
        }

        else value = stringValue.toDouble();

        if (value > item->maxValue()) item->setMaxValue(value);
        if (value < item->minValue()) item->setMinValue(value);

        newSample->appendValue(value);
    }
}

RmcPlotFile::QuantityType RmcPlotFile::quantityTypeForQuantity(QString *name, QString *formula)
{
    // regular formula pattern
    QRegularExpression regPattern(R"(^%MD(\d+)\.(\d+))");

    // rate formula pattern
    QRegularExpression ratePattern(R"(^RATE\(%MD(\d+)\.(\d+)\))");

    QRegularExpressionMatch regMatch = regPattern.match(formula);
    QRegularExpressionMatch rateMatch = ratePattern.match(formula);

    // File and element of the address
    int f, e;

    // Controller type
    RmcPlot::ControllerType cType = m_plots.last()->controllerType();

    if (regMatch.hasMatch())
    {
        f = regMatch.captured(1).toInt();
        e = regMatch.captured(2).toInt();

        // RMC200
        if (cType == RmcPlot::ControllerType::RMC200)
        {
            if (f < 256 || f > 383) return QuantityType::Other;

            if (((e == 20 || e == 32 || e == 36) && name->contains("Position")) || e == 400 || e == 404)
                return QuantityType::Position;
            else if (((e == 21 || e == 33 || e == 37) && name->contains("Velocity")) || e == 401 || e == 405)
                return QuantityType::Velocity;
            else if (e == 22 || e == 34 || e == 38 || e == 402)
                return QuantityType::Acceleration;
            else if (e == 26)
                return QuantityType::Counts;
            else if (e == 27)
                return QuantityType::RawCounts;
            else if (((e == 20 || e == 32 || e == 36) && !name->contains("Position")) || e == 80 || e == 92 || e == 96 || e == 440 || e == 444)
                return QuantityType::PressureForce;
            else if (((e == 21 || e == 33 || e == 37) && !name->contains("Velocity")) || e == 81 || e == 93 || e == 97 || e == 441)
                return QuantityType::PressureForceRate;
            else if(e == 86)
                return QuantityType::SecCounts;
            else if (e == 87)
                return QuantityType::SecRawCounts;
            else if (e == 151 || (e >= 182 && e <= 191) || (e >= 292 && e <= 301))
                return QuantityType::Gains;
            else if (e == 403)
                return QuantityType::Jerk;
            else if (e == 140)
                return QuantityType::ControlOutput200;
            return QuantityType::Other;
        }

        // RMC75 and RMC150
        else
        {
            if ((cType == RmcPlot::ControllerType::RMC70 && (f < 8 || f > 11)) || (cType == RmcPlot::ControllerType::RMC150 && (f < 8 || f > 23)))
                return QuantityType::Other;
            else if ((e == 8 && name->contains("Position")) || e == 53 || e == 56)
                return QuantityType::Position;
            else if ((e == 9 && name->contains("Velocity")) || e == 54 || e == 57)
                return QuantityType::Velocity;
            else if (e == 10 || e == 55)
                return QuantityType::Acceleration;
            else if (e == 11)
                return QuantityType::Counts;
            else if (e == 12)
                return QuantityType::RawCounts;
            else if ((e == 8 && !name->contains("Position")) || e == 23 || e == 60 || e == 61)
                return QuantityType::PressureForce;
            else if ((e == 9 && !name->contains("Velocity")) || e == 24)
                return QuantityType::PressureForceRate;
            else if (e == 26)
                return QuantityType::SecCounts;
            else if (e == 27)
                return QuantityType::SecRawCounts;
            else if ((e >= 37 && e <= 45) || (e >= 48 && e <= 52))
                return QuantityType::Gains;
            else if (e == 58)
                return QuantityType::Jerk;
            else if (e == 33)
                return QuantityType::ControlOutput75_150;
            return QuantityType::Other;
        }
    }

    else if (rateMatch.hasMatch())
    {
        f = rateMatch.captured(1).toInt();
        e = rateMatch.captured(2).toInt();

        // RMC200
        if (cType == RmcPlot::ControllerType::RMC200)
        {
            if (f < 256 || f > 383)
                return QuantityType::Other;
            if (e == 20 || e == 32 || e == 36 || e == 400)
                return QuantityType::Velocity;
            return QuantityType::Other;
        }

        // RMC75/RMC150
        else
        {
            if ((cType == RmcPlot::ControllerType::RMC70 && (f < 8 || f > 11)) || (cType == RmcPlot::ControllerType::RMC150 && (f < 8 || f > 23)))
                return QuantityType::Other;
            if (e == 8 || e == 53)
                return QuantityType::Velocity;
            return QuantityType::Other;
        }
    }

    return QuantityType::Other;
}

bool RmcPlotFile::parseBool(const QString *attribute, bool *destination, bool defaultValue, QXmlStreamAttributes *attributes)
{
    if (attributes->hasAttribute(*attribute))
    {
        *destination = attributes->value(*attribute).compare(QString("True")) == 0;
        return true;
    }
    else
    {
        *destination = defaultValue;
        return false;
    }
}

bool RmcPlotFile::parseReal(const QString *attribute, qreal *destination, qreal defaultValue, QXmlStreamAttributes *attributes)
{
    if (attributes->hasAttribute(*attribute))
    {
        *destination = attributes->value(*attribute).toDouble();
        return true;
    }
    else
    {
        *destination = defaultValue;
        return false;
    }
}

bool RmcPlotFile::parseInt(const QString *attribute, int *destination, int defaultValue, QXmlStreamAttributes *attributes)
{
    if (attributes->hasAttribute(*attribute))
    {
        *destination = attributes->value(*attribute).toInt();
        return true;
    }
    else
    {
        *destination = defaultValue;
        return false;
    }
}

void RmcPlotFile::getQuantitiesAndValues(qreal width, qreal height, RmcPlot* data, QList<QuantityValuePair>* quantitiesAndValues)
{
    // Find range of intervals recorded.
    // In this case, we're getting the whole range, but in the future, we may be able to restrict it to what's visible.
    const int first = data->samples().firstKey();
    const int last = data->samples().lastKey();
    const int range = last - first;

    // Determine how many intervals a single pixel column should hold
    const qreal intsPerColumn = qMax(1., (qreal)range / (qreal)width);

    // Create vector holding bounds for each column
    auto groups = QList<QPair<qreal, qreal>>();
    for (int i = 0; i < width; i++)
    {
        qreal left = first + i * intsPerColumn;
        qreal right = first + (i + 1) * intsPerColumn;
        groups.append(QPair<double, double>(left, right));
    }

    // Update scale groups for auto-scaling and center-on-zero behavior
    updateScaleGroups(data);

    // Scale groups all set up now!
    // Next, let's go through the quantities, and get their pixel column values
    populateQuantityPixelColumns(data, &groups, first, range, width, height, quantitiesAndValues);
}


void RmcPlotFile::getQuantitiesAndValuesXY(qreal width, qreal height, RmcPlot* data, QList<QuantityValuePair> *quantitiesAndValues, int xQuantityIndex)
{
    const int first = data->samples().firstKey();
    const int last = data->samples().lastKey();
    const int range = last - first;

    const qreal intsPerColumn = qMax(1., (qreal)range / (qreal)width);

    auto groups = QList<QPair<qreal, qreal>>();
    for (int i = 0; i < width; i++)
    {
        qreal left = first + i * intsPerColumn;
        qreal right = first + (i + 1) * intsPerColumn;
        groups.append(QPair<double, double>(left, right));
    }

    updateScaleGroups(data);

    populateQuantityPixelColumnsXY(data, &groups, first, range, width, height, data->quantities().at(xQuantityIndex), quantitiesAndValues);
}

void RmcPlotFile::updateScaleGroups(RmcPlot* data)
{
    for (auto i = data->scaleGroups().begin(); i != data->scaleGroups().end(); ++i)
    {
        int id = i.key();
        RmcScaleGroup* group = i.value();

        if (group->autoScale())
        {
            // Locate quantities with this scale group
            for (auto& quantity : data->quantities())
            {
                if (quantity->scaleGroup() != id) continue;

                // Find plot item that this quantity corresponds to
                QString formula = quantity->formula();

                // Now we need to get corresponding plot item
                for (auto item : data->items())
                {
                    if (item->address() != formula) continue;

                    // Check if item's bounds should replace scale group's bounds
                    if (item->minValue() < group->bottomScale()) group->setBottomScale(item->minValue());
                    if (item->maxValue() > group->topScale()) group->setTopScale(item->maxValue());
                }
            }

            // Center on zero
            if (group->centerOn0())
            {
                qreal bottom = group->bottomScale();
                qreal top = group->topScale();

                if (abs(bottom) > abs(top))
                    group->setTopScale(bottom * -1);
                else
                    group->setBottomScale(top * -1);
            }
        }

        // Handle minimum range
        // Default range is 1
        // NOTE: If TopScale or BottomScale were explicitly defined, don't adjust them
        qreal currentRange = group->topScale() - group->bottomScale();
        if (currentRange < group->minRange())
        {
            if (!group->isBottomScaleExplicit() && !group->isTopScaleExplicit())
            {
                qreal toAdd = group->minRange() / 2.0f;

                // If this group should be centered on zero, then center it on zero!
                if (group->centerOn0())
                {
                    group->setTopScale(toAdd);
                    group->setBottomScale(-toAdd);
                }

                // Otherwise, center it around the average of the current range
                else
                {
                    qreal average = (group->bottomScale() + group->topScale()) / 2.0f;
                    group->setTopScale(average + toAdd);
                    group->setBottomScale(average - toAdd);
                }
            }
            else if (!group->isBottomScaleExplicit())
            {
                group->setBottomScale(group->topScale() - group->minRange());
            }
            else if (!group->isTopScaleExplicit())
            {
                group->setTopScale(group->bottomScale() + group->minRange());
            }
        }
    }
}

void RmcPlotFile::populateQuantityPixelColumns(RmcPlot* data,
                                               QList<QPair<qreal, qreal>>* groups,
                                               int firstInt, int intRange,
                                               qreal width, qreal height,
                                               QList<QuantityValuePair>* quantitiesAndValues)
{
    for (auto& quantity : data->quantities())
    {
        // Find what item this corresponds to
        int index = -1;
        for (int i = 0; i < data->items().size(); i++)
        {
            auto item = data->items().at(i);

            auto itemAddress = item->address();
            auto quantityFormula = quantity->formula();

            if (itemAddress == quantityFormula)
            {
                index = i;
                break;
            }
        }

        // Skip if not graphable
        if (index == -1) continue;

        const auto& scaleGroup = data->scaleGroups().value(quantity->scaleGroup());

        auto columns = new QList<RmcPlotPixelColumn>();

        for (auto& bounds : qAsConst(*groups))
        {
            qreal left = bounds.first;
            qreal right = bounds.second;

            qreal first = NAN, highest = DBL_MIN, lowest = DBL_MAX, last = NAN;

            // Iterate through samples from left to right bound
            bool firstSet = false;
            for (int i = (int)left; i < (int)right; i++)
            {
                if (data->samples().count(i) == 0) continue;

                const auto& thisSample = data->samples().value(i);

                if (thisSample == nullptr)
                    continue;

                qreal currentValue = thisSample->valueAt(index);

                // Only assign to first value once (the first time)
                if (!firstSet)
                {
                    firstSet = true;
                    first = currentValue;
                    highest = currentValue;
                    lowest = currentValue;
                }

                // Assign to last value every time, that way it'll always be the last read value.
                last = currentValue;

                // Assign highest and lowest values.
                if (highest < currentValue) highest = currentValue;
                if (lowest > currentValue) lowest = currentValue;
            }

            // X position of column is given by ratio of left to lastInt, multiplied by height
            qreal x = (width * ((qreal)(left - firstInt) / (qreal)intRange));
            if (isnan(first) || isnan(last)) continue;

            // Store data in pixel column structure
            auto column = RmcPlotPixelColumn(first, highest, lowest, last, scaleGroup, height, x);
            columns->append(column);
        }

        quantitiesAndValues->append(QuantityValuePair(quantity, columns));
    }
}

void RmcPlotFile::populateQuantityPixelColumnsXY(RmcPlot *data,
                                                 QList<QPair<qreal, qreal>> *groups,
                                                 int firstInt, int intRange,
                                                 qreal width, qreal height,
                                                 RmcPlotQuantity* xQuantity,
                                                 QList<QuantityValuePair> *quantitiesAndValues)
{

    // Get item index of X axis quantity
    int xIndex = -1;
    for (int i = 0; i < data->items().size(); i++)
    {
        auto& item = data->items().at(i);
        auto itemAddress = item->address();
        auto quantityFormula = xQuantity->formula();

        if (itemAddress == quantityFormula)
        {
            xIndex = i;
            break;
        }
    }

    if (xIndex == -1)
    {
        qDebug() << "populateQuantityPixelColumnsXY - couldn't find corresponding item for provided quantity";
        return;
    }

    // Get left and right bounds of X axis quantity
    const auto xScaleGroup = data->scaleGroups().value(xQuantity->scaleGroup());
    qreal xBottomScale = xScaleGroup->bottomScale();
    qreal xTopScale = xScaleGroup->topScale();



    for (auto& quantity : data->quantities())
    {
        // First, we need to find the item this quantity corresponds to.
        // If it's not an item we can graph, we'll skip it.
        int index = -1;
        for (int i = 0; i < data->items().size(); i++)
        {
            auto item = data->items().at(i);
            auto itemAddress = item->address();
            auto quantityFormula = quantity->formula();

            if (itemAddress == quantityFormula)
            {
                // This quantity corresponds to this item!
                index = i;
                break;
            }
        }

        if (index == -1) continue; // If this quantity didn't have an associated item, skip it.


        // The scale group for this quantity.
        auto scaleGroup = data->scaleGroups().value(quantity->scaleGroup());

        auto columns = new QList<RmcPlotPixelColumn>();

        // This is where this function should differ from populateQuantityPixelColumns() -
        // for an XY plot, X should be the value of the X axis quantity at this time!
        qreal x = 0;

        for (auto& bounds : qAsConst(*groups))
        {
            qreal leftBound = bounds.first;
            qreal rightBound = bounds.second;

            qreal first = NAN, highest = DBL_MIN, lowest = DBL_MAX, last = NAN;

            bool firstSet = false;
            for (int i = (int)leftBound; i < (int)rightBound; i++)
            {
                if (data->samples().count(i) == 0) continue;

                auto thisSample = data->samples().value(i);

                qreal currentValue = thisSample->valueAt(index);

                // Only assign to the first value once (the first time)
                if (!firstSet)
                {
                    firstSet = true;
                    first = currentValue;
                    highest = currentValue;
                    lowest = currentValue;

                    x = thisSample->valueAt(xIndex);

                }


                // Assign to last value every time, that way it'll always be the last read value
                last = currentValue;

                // Assign highest and lowest values
                if (highest < currentValue) highest = currentValue;
                if (lowest > currentValue) lowest = currentValue;
            }

            // Normalize x
            x = remap(xBottomScale, xTopScale, 0, width, x);

            // Store data in pixel column structure
            auto column = RmcPlotPixelColumn(first, highest, lowest, last, scaleGroup, height, x);
            columns->append(column);
        }



        quantitiesAndValues->append(QuantityValuePair(quantity, columns));
    }
}

QImage* RmcPlotFile::getXYImage(qreal width, qreal height, int plotIndex, int xQuantityIndex)
{
    // Initialize the image and painter
    auto image = new QImage((int)width, (int)height, QImage::Format::Format_ARGB32_Premultiplied);
    image->fill(QColorConstants::Transparent);
    QPainter painter(image);
    painter.setRenderHint(QPainter::RenderHint::Antialiasing, true);

    if (plots().count() == 0)
    {
        qDebug() << "getXYImage() - there are no plots to load";
        return image;
    }

    auto plot = plots().at(plotIndex);
    auto xQuantity = plot->quantities().at(xQuantityIndex);

    // Get the item index of the X axis quantity
    int xIndex = -1;
    for (int i = 0; i < plot->items().size(); i++)
    {
        auto item = plot->items().at(i);
        auto itemAddress = item->address();
        if (itemAddress == xQuantity->formula())
        {
            xIndex = i;
            break;
        }
    }

    if (xIndex == -1)
    {
        qDebug() << "getXYImage - couldn't find corresponding item index for X axis quantity";
        return nullptr;
    }

    // Get the left and right bounds of the X axis quantity
    auto xScaleGroup = plot->scaleGroups().value(xQuantity->scaleGroup());
    qreal xBottomScale = xScaleGroup->bottomScale();
    qreal xTopScale = xScaleGroup->topScale();

    // Iterate over all of the quantities, and draw them (or, queue them to be drawn)
    for (auto& quantity : plot->quantities())
    {
        // Don't graph if we can't find the item that this quantity corresponds to,
        // or if we're not supposed to graph it

        if (!quantity->graphQuantity() || !quantity->userGraph()) continue;

        int yIndex = -1;
        for (int i = 0; i < plot->items().size(); i++)
        {
            auto item = plot->items().at(i);
            auto itemAddress = item->address();
            if (itemAddress == quantity->formula())
            {
                yIndex = i;
                break;
            }
        }

        if (yIndex == -1) continue;

        // If we've gotten to this point, then we know we want to graph this quantity.

        auto yScaleGroup = plot->scaleGroups().value(quantity->scaleGroup());
        qreal yBottomScale = yScaleGroup->bottomScale();
        qreal yTopScale = yScaleGroup->topScale();

        // True height depends on whether or not the quantity is rail-to-rail
        qreal imageTop = yScaleGroup->railToRail() ? 0 : (height * (1.0 - RAIL_TO_RAIL_RATIO));
        qreal imageBottom = yScaleGroup->railToRail() ? height : (height * RAIL_TO_RAIL_RATIO);

        const int firstSample = plot->samples().firstKey();
        const int lastSample = plot->samples().lastKey();
        QPolygonF polyline;
        for (int i = firstSample; i < lastSample; i++)
        {
            if (plot->samples().count(i) == 0) continue;
            auto currentSample = plot->samples().value(i);

            // Determine the current X value
            qreal currentXValue = currentSample->valueAt(xIndex);

            // Determine the current Y value
            qreal currentYValue = currentSample->valueAt(yIndex);

            // Remap the X and Y values to image space
            qreal remappedXValue = remap(xBottomScale, xTopScale, 0, width, currentXValue);
            qreal remappedYValue = remap(yTopScale, yBottomScale, imageTop, imageBottom, currentYValue);

            polyline << QPointF(remappedXValue, remappedYValue);
        }

        painter.setBrush(Qt::NoBrush);
        painter.setPen(quantity->color());
        painter.drawPolyline(polyline);
    }
    return image;
}

QImage* RmcPlotFile::renderXYImage(qreal width, qreal height, QList<QuantityValuePair> *quantitiesAndValues)
{
    QImage *img = new QImage((int)width, (int)height, QImage::Format::Format_ARGB32_Premultiplied);
    img->fill(QColorConstants::Transparent);

    if (img->isNull())
    {
        return img;
    }

    QPainter painter(img);
    painter.setRenderHint(QPainter::RenderHint::Antialiasing, true);

    for (auto quantityValuePair : qAsConst(*quantitiesAndValues))
    {
        paintQuantity(&quantityValuePair, &painter);
    }

    return img;
}

QImage* RmcPlotFile::getTYImage(qreal width, qreal height, int plotIndex)
{
    QList<QuantityValuePair> quantitiesAndValues;
    getQuantitiesAndValues(width, height, plots().at(plotIndex), &quantitiesAndValues);
    return renderImage(width, height, &quantitiesAndValues);
}

/*
 * We only need to re-render the image when:
 * 1) the window is resized
 * 2) the bounds are resized
 */
QImage* RmcPlotFile::renderImage(qreal width, qreal height, QList<QuantityValuePair> *quantitiesAndValues)
{
    QImage* img = new QImage((int)width, (int)height, QImage::Format::Format_ARGB32_Premultiplied);
    img->fill(QColorConstants::Transparent);

    if (img->isNull())
    {
        return img;
    }

    QPainter painter(img);
    painter.setRenderHint(QPainter::RenderHint::Antialiasing, true);

    for (auto quantityValuePair : qAsConst(*quantitiesAndValues))
    {
        paintQuantity(&quantityValuePair, &painter);
    }

    return img;
}

void RmcPlotFile::paintQuantity(QuantityValuePair *quantityAndColumns, QPainter *painter)
{
    auto quantity = quantityAndColumns->first;
    auto columns = quantityAndColumns->second;

    if (!quantity->graphQuantity() || !quantity->userGraph()) return;

    QColor color = quantity->color();

    painter->setPen(color);

    QVector<QPointF> allPoints;

    for (auto thisColumn : qAsConst(*columns))
    {
        if (isnan(thisColumn.first()) || isnan(thisColumn.last())) continue;

        allPoints.append(QPointF(thisColumn.x(), thisColumn.first()));
        allPoints.append(QPointF(thisColumn.x(), thisColumn.highest()));
        allPoints.append(QPointF(thisColumn.x(), thisColumn.lowest()));
        allPoints.append(QPointF(thisColumn.x(), thisColumn.last()));
    }
    painter->drawPolyline(QPolygonF(allPoints));
}
