#include "plotmath.h"

std::tuple<qreal, int> PlotMath::computeUnitsPerGrid(int cyPerInch, qreal dScale)
{
    if (dScale <= 0.0)
    {
        qDebug() << "computeUnitsPerGrid - dScale should be > 0.0, but it is" << dScale;
        return std::tuple<float, int>{0, 0};
    }

    GridMultiple nState = GridMultiple::DIV_1;

    qreal dBase;
    qreal dLGridSize;
    qreal dUnitsPerGridLast;
    qreal dLGridSizeLast;
    qreal nDecPlLast;
    qreal dOptLGridSize = 20.;

    // Start by assuming 1 unit per grid, which has no decimal places, and a base of 1.
    qreal dUnitsPerGrid = 1.;
    int nDecPl = 0;
    dBase = 1.;

    dScale = 1. / (cyPerInch * dScale);

    dLGridSize = dUnitsPerGrid * dScale;

    if (dLGridSize < dOptLGridSize)
    {
        do
        {
            dUnitsPerGridLast = dUnitsPerGrid;
            dLGridSizeLast = dLGridSize;

            switch (nState)
            {
            case GridMultiple::DIV_1:
                nState = GridMultiple::DIV_2;
                dUnitsPerGrid = dBase * 2.;
                break;
            case GridMultiple::DIV_2:
                nState = GridMultiple::DIV_5;
                dUnitsPerGrid = dBase * 5.;
                break;
            case GridMultiple::DIV_5:
                nState = GridMultiple::DIV_1;
                dBase *= 10.0;
                dUnitsPerGrid = dBase;
                break;
            }

            dLGridSize = dUnitsPerGrid * dScale;
        }
        while (dLGridSize < dOptLGridSize);

        if (dLGridSize - dOptLGridSize > dOptLGridSize - dLGridSizeLast)
            dUnitsPerGrid = dUnitsPerGridLast;
    }

    else
    {
        do
        {
            dUnitsPerGridLast = dUnitsPerGrid;
            dLGridSizeLast = dLGridSize;
            nDecPlLast = nDecPl;

            switch (nState)
            {
            case GridMultiple::DIV_1:
                nState = GridMultiple::DIV_5;
                dBase *= 10.0;
                nDecPl++;
                dUnitsPerGrid = 5. / dBase;
                break;
            case GridMultiple::DIV_2:
                nState = GridMultiple::DIV_1;
                dUnitsPerGrid = 1. / dBase;
                break;
            case GridMultiple::DIV_5:
                nState = GridMultiple::DIV_2;
                dUnitsPerGrid = 2. / dBase;
                break;
            }

            dLGridSize = dUnitsPerGrid * dScale;
        }
        while (dLGridSize > dOptLGridSize);

        if (dOptLGridSize - dLGridSize > dLGridSizeLast - dOptLGridSize)
        {
            dUnitsPerGrid = dUnitsPerGridLast;
            nDecPl = nDecPlLast;
        }
    }

    dUnitsPerGrid /= 5.0;

    return std::tuple<float, int>{dUnitsPerGrid, nDecPl};
}

std::tuple<qreal, qreal, int, qreal> PlotMath::computeTimeScale(qreal dIntervalsPerPixel, int cyPerInch,
                                                             qreal samplePeriod, int leftmostInt,
                                                             int firstIntEver)
{
    qreal dTScale = dIntervalsPerPixel * samplePeriod;

    qreal dTUnitsPerGrid;
    int nTDecPl;
    std::tie(dTUnitsPerGrid, nTDecPl) = PlotMath::computeUnitsPerGrid(cyPerInch, dTScale);

    qreal dTValAtP0 = (qreal)(leftmostInt - firstIntEver) * samplePeriod;

    return std::tuple<qreal, qreal, int, qreal>{dTScale, dTUnitsPerGrid, nTDecPl, dTValAtP0};

}

TickMarkData PlotMath::calculateTickMarks(RmcPlotQuantity *quantity, bool allowPadding, qreal cyPerInch, qreal physicalSize)
{
    auto plot = Engine::engine->currentPlot();
    if (plot == nullptr || quantity == nullptr)
    {
        return {false};
    }

    auto scaleGroupIndex = quantity->scaleGroup();
    auto scaleGroup = plot->scaleGroups().value(scaleGroupIndex);

    qreal bottomBound = scaleGroup->bottomScale();
    qreal topBound = scaleGroup->topScale();

    // This code calculates the step size for the vertical tick marks.
    // The computeUnitsPerGrid() function is translated from RMCTools
    // code.
    qreal range = topBound - bottomBound;
    qreal scale = range / physicalSize;

    if (range <= 0.0 || scale <= 0.0)
    {
        return {false};
    }
    qreal unitsPerGrid;
    int decPlaces;

    std::tie(unitsPerGrid, decPlaces) = PlotMath::computeUnitsPerGrid(cyPerInch, scale);

    // Add padding to the bounds if necessary. Despite the solution
    // being pretty simple, this took me a while to get working correctly.

    // Bonus Tip: Always make sure your padding values are consistent for the
    // lower and upper bounds! I'd been multiplying the bottom and top
    // bounds by the ratio amount, which wasn't working because they would
    // scale differently.
    if (allowPadding && !scaleGroup->railToRail())
    {
        qreal padding = range * (1 - RAIL_TO_RAIL_RATIO);
        bottomBound -= padding;
        topBound += padding;
    }

    // Here, we need to find the location of our first tick mark. We start
    // at zero, then move in increments of our tick mark towards the bottom
    // bound (either + or - ) until we reach the bottom bound.
    qreal tickLocation = 0;
    if (bottomBound >= 0)
        while (tickLocation < bottomBound) tickLocation += unitsPerGrid;
    else
        while (tickLocation > bottomBound) tickLocation -= unitsPerGrid;

    TickMarkData data = {
        true,
        unitsPerGrid,
        decPlaces,
        bottomBound,
        topBound,

        tickLocation
    };

    return data;
}
