#include "mainwindow.h"
#include <QBuffer>
#include <QFileDialog>

#include "Models/Engine/engine.h"

MainWindow::MainWindow(QWidget *parent)
    : QMainWindow(parent)
{
    m_videoView = new VideoView(this);
    m_controlView = new ControlView(this);
    m_timelineView = new TimelineView(this);

    m_syncAction = new QAction("Synchronize", this);
    m_syncAction->setIcon(QIcon(":/icons/loop"));
    m_syncAction->setCheckable(true);
    m_syncAction->setShortcut(QKeySequence("Ctrl+1"));
    connect(m_syncAction, &QAction::triggered, m_timelineView, &TimelineView::setSynchronizeMode);

    m_analyzeAction = new QAction("Analyze", this);
    m_analyzeAction->setIcon(QIcon(":/icons/ty"));
    m_analyzeAction->setCheckable(true);
    m_analyzeAction->setShortcut(QKeySequence("Ctrl+2"));
    connect(m_analyzeAction, &QAction::triggered, m_timelineView, &TimelineView::setAnalyzeMode);

    m_workspaceActionGroup = new QActionGroup(this);
    m_workspaceActionGroup->addAction(m_syncAction);
    m_workspaceActionGroup->addAction(m_analyzeAction);

    connect(Engine::engine, &Engine::invalidTimelineItemLoaded,
        [=](TimelineItem* item)
        {
            QMessageBox::critical(this, "Invalid video", "This video could not be loaded.");
            Engine::engine->deleteTimelineItem(item);
        });

    connect(Engine::engine, &Engine::plotLoadingFailed, [=](auto errorString)
    {
        QMessageBox::critical(this, "Invalid plot", QString("The plot could not be loaded:\n%1").arg(errorString));
    });

    // Set up properties window
    m_propertiesDockWidget = new QDockWidget("Properties", this);
    m_propertiesView = new PropertiesView(this);
    m_propertiesDockWidget->setWidget(m_propertiesView);
    addDockWidget(Qt::DockWidgetArea::RightDockWidgetArea, m_propertiesDockWidget);

    // Set up quantities window
    m_quantityPanelDockWidget = new QDockWidget("Quantities", this);
    m_quantityPanelView = new QuantityPanelView(this);
    m_quantityPanelDockWidget->setWidget(m_quantityPanelView);
    addDockWidget(Qt::DockWidgetArea::LeftDockWidgetArea, m_quantityPanelDockWidget);

    // Set up XY plot window
    m_xyPlotPanelDockWidget = new QDockWidget("XY Plot", this);
    m_xyPlotPanelView = new XYPlotPanelView(this);
    m_xyPlotPanelDockWidget->setWidget(m_xyPlotPanelView);
    addDockWidget(Qt::DockWidgetArea::RightDockWidgetArea, m_xyPlotPanelDockWidget);

    // File menu actions
    m_openAction = new QAction("&Open Configuration...", this);
    m_openAction->setShortcut(QKeySequence::Open);
    connect(m_openAction, &QAction::triggered, this, &MainWindow::showOpenModal);

    m_saveAction = new QAction("&Save Configuration...", this);
    m_saveAction->setShortcut(QKeySequence::Save);
    connect(m_saveAction, &QAction::triggered, this, &MainWindow::showSaveAsModal);

    m_loadVideoAction = new QAction("&Import Video...", this);
    connect(m_loadVideoAction, &QAction::triggered, this, &MainWindow::showImportVideoModal);

    m_loadBaseVideoAction = new QAction("&Set Base Video...", this);
    connect(m_loadBaseVideoAction, &QAction::triggered, this, &MainWindow::showImportBaseVideoModal);

    m_loadPlotAction = new QAction("&Set Plot...", this);
    connect(m_loadPlotAction, &QAction::triggered, this, &MainWindow::showImportPlotModal);

    m_exitAction = new QAction("&Quit", this);
    m_exitAction->setShortcut(QKeySequence::Quit);
    m_exitAction->setMenuRole(QAction::MenuRole::QuitRole);
    connect(m_exitAction, &QAction::triggered, QApplication::quit);


    // Help menu actions
    m_openHelpAction = new QAction("&Open Help...", this);
    m_openHelpAction->setShortcut(QKeySequence::HelpContents);
    connect(m_openHelpAction, &QAction::triggered, this, &MainWindow::showHelpWindow);

    m_submitBugReportAction = new QAction("&Submit Bug Report...", this);
    m_submitBugReportAction->setEnabled(false);
    connect(m_submitBugReportAction, &QAction::triggered, this, [=]() {
        QMessageBox::critical(this, "VideoSync", "The bug report submitter could not be launched. Please report this as a bug.");
    });

    m_aboutVideoSyncAction = new QAction("&About VideoSync...", this);
    m_aboutVideoSyncAction->setMenuRole(QAction::MenuRole::AboutRole);
    connect(m_aboutVideoSyncAction, &QAction::triggered, this, [=]()
    {
        QMessageBox::about(this, "VideoSync",
                           "<b>VideoSync</b><br>"
                           "Programmed by Malcolm Anderson<br>"
                           "Copyright (c) Delta Computer Systems 2022");
    });

    m_aboutQtAction = new QAction("&About Qt...", this);
    m_aboutQtAction->setMenuRole(QAction::MenuRole::AboutQtRole);
    connect(m_aboutQtAction, &QAction::triggered, this, [=]()
    {
        QMessageBox::aboutQt(this, "About Qt");
    });


    // Set up menu bar
    m_fileMenu = menuBar()->addMenu("&File");
    m_fileMenu->addAction(m_openAction);
    m_fileMenu->addAction(m_saveAction);
    m_fileMenu->addSeparator();
    m_fileMenu->addAction(m_loadVideoAction);
    m_fileMenu->addAction(m_loadBaseVideoAction);
    m_fileMenu->addAction(m_loadPlotAction);
    m_fileMenu->addSeparator();
    m_fileMenu->addAction(m_exitAction);

    m_viewMenu = menuBar()->addMenu("&View");
    m_viewMenu->addAction(&Engine::engine->zoomOutAction());
    m_viewMenu->addAction(&Engine::engine->zoomInAction());

    m_workspaceMenu = menuBar()->addMenu("&Workspace");
    m_workspaceMenu->addAction(m_syncAction);
    m_workspaceMenu->addAction(m_analyzeAction);

    m_windowMenu = menuBar()->addMenu("&Window");
    m_windowMenu->addAction(m_propertiesDockWidget->toggleViewAction());
    m_windowMenu->addAction(m_quantityPanelDockWidget->toggleViewAction());
    m_windowMenu->addAction(m_xyPlotPanelDockWidget->toggleViewAction());

    m_helpMenu = menuBar()->addMenu("&Help");
    m_helpMenu->addAction(m_openHelpAction);
    m_helpMenu->addAction(m_submitBugReportAction);
    m_helpMenu->addSeparator();
    m_helpMenu->addAction(m_aboutVideoSyncAction);
    m_helpMenu->addAction(m_aboutQtAction);

    // Upper layout
    m_upperWidget = new QWidget(this);
    m_upperLayout = new QVBoxLayout(m_upperWidget);

    m_upperLayout->addWidget(m_videoView);
    m_upperLayout->addWidget(m_controlView);

    m_mainSplitter = new QSplitter(Qt::Orientation::Vertical, this);
    m_mainSplitter->setChildrenCollapsible(false);
    m_mainSplitter->addWidget(m_upperWidget);
    m_mainSplitter->addWidget(m_timelineView);

    m_overallWidget = new QWidget(this);
    m_overallLayout = new QVBoxLayout(m_overallWidget);
    m_overallLayout->addWidget(m_mainSplitter);
    setCentralWidget(m_overallWidget);

    // Make the background of QVideoWidgets black
    setStyleSheet("QVideoWidget { background-color: black }");

    // By default, load into synchronize mode
    m_syncAction->trigger();
}

void MainWindow::showImportVideoModal()
{
    QStringList videos = QFileDialog::getOpenFileNames(
                this,
                "Import Video",
                "",
                "Videos (*.mp4 *.avi *.mov)");

    if (videos.isEmpty()) return;

    // Add the videos to the engine
    for (auto& path : videos)
    {
        auto timelineItem = new TimelineItem(this);
        timelineItem->setOffset(0.0);

        // Determine the name of the video
        QFileInfo fileInfo(path);
        QString filename = fileInfo.baseName();
        timelineItem->setName(filename);
        timelineItem->loadVideo(path);
        Engine::engine->addTimelineItem(timelineItem);
    }
}

void MainWindow::showImportBaseVideoModal()
{
    QString path = QFileDialog::getOpenFileName(
                this,
                "Import Base Video",
                "",
                "Videos (*.mp4 *.avi *.mov)");

    if (path.isNull()) return;

    // Add the videos to the engine
    if (!path.isEmpty())
    {
        auto timelineItem = new TimelineItem(this);
        timelineItem->setOffset(0.0);

        // Determine the name of the video
        QFileInfo fileInfo(path);
        QString filename = fileInfo.baseName();
        timelineItem->setName(filename);
        timelineItem->loadVideo(path);

        Engine::engine->baseTimelineItem()->deleteLater();
        Engine::engine->setBaseTimelineItem(timelineItem);
    }
}

void MainWindow::showImportPlotModal()
{
    qDebug() << "import plot time";
    QString path = QFileDialog::getOpenFileName(
                this,
                "Import Plot",
                "",
                "RMCTools Plot Files (*.rmcplots *.rmcplotx)");

    if (path.isNull()) return;

    if (!path.isEmpty())
    {
        Engine::engine->loadPlotFile(path);
    }
}

void MainWindow::showSaveAsModal()
{
    QString dest = QFileDialog::getSaveFileName(
                this,
                "Save File",
                "",
                "VideoSync Configuration (*.vsync)");

    if (dest.isNull()) return;

    qDebug() << "Save as" << dest;

    QFile f(dest);
    if (!f.open(QIODevice::WriteOnly))
    {
        QMessageBox::critical(this, "VideoSync",
                              QString("The file could not be saved: %1").arg(f.errorString())
                              );
        return;
    }

    Engine::engine->writeXml(&f);
    f.close();
}

void MainWindow::showOpenModal()
{
    QString path = QFileDialog::getOpenFileName(
                this,
                "Open Configuration",
                "",
                "VideoSync Configuration (*.vsync *.xml)");

    if (path.isNull()) return;

    QFile f(path);
    if (!f.open(QIODevice::ReadOnly))
    {
        QMessageBox::critical(this, "VideoSync",
                              QString("The file could not be opened: %1").arg(f.errorString())
                              );
        return;
    }

    Engine::engine->readXml(&f);
    f.close();
}

void MainWindow::showHelpWindow()
{
    // The window should be destroyed, and its memory freed, when it's closed
    HelpWindow* helpWindow = new HelpWindow(this);
    helpWindow->show();
}
