#include "timelineview.h"
/*!
 * \class TimelineView
 * \brief The TimelineView class is a really cool class
 *
 * TimelineView displays the Timeline, which may include a plot
 * and video clips.
 */

TimelineView::TimelineView(QWidget *parent) : QWidget(parent)
{
    m_plotDrawer = new PlotDrawer(this);
    m_redrawPlotTimer = new QTimer(this);
    setFocusPolicy(Qt::StrongFocus);
    setSizePolicy(QSizePolicy::Policy::Expanding, QSizePolicy::Policy::Minimum);
    connect(Engine::engine, &Engine::positionChanged, this, &TimelineView::onPositionChanged);
    connect(Engine::engine, &Engine::zoomBoundsChanged, this, [=]() { invalidatePlot(); update(); });
    // Set up lock renderer
    m_lockFillRenderer.loadSvg(":/icons/lock");
    m_lockDisabledRenderer.loadSvg(":/icons/lock_no");

    m_visibilityFillRenderer.loadSvg(":/icons/eye");
    m_visibilityDisabledRenderer.loadSvg(":/icons/eye_no");

    m_earthRenderer.loadSvg(":/icons/earth");

    connect(Engine::engine, &Engine::plotVisibilityChanged, this, &TimelineView::onRedrawTimerExpired);
    connect(m_redrawPlotTimer, &QTimer::timeout, this, &TimelineView::onRedrawTimerExpired);
    m_redrawPlotTimer->setSingleShot(true);

    connect(m_plotDrawer, &PlotDrawer::plotsComplete, this, &TimelineView::onRedrawPlotsComplete);
    connect(Engine::engine, &Engine::plotLoaded, this, &TimelineView::onRedrawTimerExpired);

    // Set up the connections for right-click context menus.
    setContextMenuPolicy(Qt::ContextMenuPolicy::CustomContextMenu);
    connect(this, &QWidget::customContextMenuRequested, this, &TimelineView::onContextMenuRequested);

    connect(Engine::engine, &Engine::plotYQuantityIndexChanged, this, [=]() { update(); });

    connect(Engine::engine, &Engine::timingsChanged, this, [=]() { update(); });
}

void TimelineView::onRedrawPlotsComplete(QImage* plotImage)
{
    if (m_plotImage != nullptr) delete m_plotImage;
    m_plotImage = plotImage;
    update();
}

void TimelineView::onRedrawTimerExpired()
{
    if (Engine::engine->plotFile() == nullptr)
    {
        // no plot file is loaded
        return;
    }

    if (Engine::engine->plotFile()->plots().count() == 0)
    {
        // onRedrawTimerExpired is running, but there are no plots loaded!
        return;
    }
    QRectF plotItemPreviewRect = getPlotItemRect();
    plotItemPreviewRect.setTop(plotItemPreviewRect.top() + videoRowHeight * analyzeAnimationAmount);

    // Set up plot drawing thread
    m_plotDrawer->setSize(QSize(plotItemPreviewRect.width(), plotItemPreviewRect.height()));
    m_plotDrawer->start();
}

bool TimelineView::event(QEvent* event)
{
    if (event->type() == QEvent::ToolTip)
    {
        toolTipEvent(static_cast<QHelpEvent*>(event));
        return true;
    }
    return QWidget::event(event);
}

void TimelineView::toolTipEvent(QHelpEvent* event)
{
    QString toolTipText = "";

    // Check if it's in the plot's lock box
    if (getPlotLockButtonRect().contains(event->pos()))
    {
        toolTipText = QString("Plot %1")
                .arg(Engine::engine->plotLocked() ? "locked" : "unlocked");
    }

    // Check if it's in the plot's visibility box
    else if (getPlotVisibilityButtonRect().contains(event->pos()))
    {
        toolTipText = "Plot visiblity cannot be changed";
    }

    // Check if it's in the plot itself
    else if (getPlotItemRect().contains(event->pos()))
    {
        toolTipText = QString("Plot: \"%1\"")
                .arg(Engine::engine->plotFile()->name());
    }

    // Check if it's in any of the timeline items
    else
    {
        for (int i = 0; i < Engine::engine->timelineItems().size() + 1; i++)
        {
            auto timelineItem = Engine::engine->timelineItem(i);
            if (getVideoLockButtonRect(i).contains(event->pos()))
            {
                if (timelineItem->isBaseLayer())
                {
                    toolTipText = QString("Base video cannot be unlocked");
                }
                else
                {
                    toolTipText = QString("Video %1")
                            .arg(timelineItem->locked() ? "locked" : "unlocked");
                }
            }
            else if (getVideoVisibilityButtonRect(i).contains(event->pos()))
            {
                toolTipText = QString("%1ideo %2")
                        .arg(
                            (timelineItem->isBaseLayer() ? "Base v" : "V"),
                            (timelineItem->visible() ? "visible" : "hidden")
                            );
            }
            else if (getVideoItemRect(i).contains(event->pos()))
            {
                toolTipText = QString("%1ideo: \"%2\" (%3s)")
                        .arg(
                            (timelineItem->isBaseLayer() ? "Base v" : "V"),
                            timelineItem->name(),
                            QString::number((qreal)timelineItem->mediaPlayer()->duration() / 1000.0)
                            );
            }
        }
    }


    // Lastly, assign the tool tip text if it exists
    if (toolTipText.isEmpty())
    {
        QToolTip::hideText();
    }
    else
    {
        QToolTip::showText(event->globalPos(), toolTipText);
    }

}
void TimelineView::resizeEvent(QResizeEvent *event)
{
    QWidget::resizeEvent(event);
    invalidatePlot();
}

void TimelineView::invalidatePlot()
{
    m_redrawPlotTimer->start(100);
}

void TimelineView::onPositionChanged() { update(); }

QSize TimelineView::sizeHint() const { return QSize(0, 150); }

void TimelineView::paintEvent(QPaintEvent */*event*/)
{
    QPainter painter(this);
    painter.setRenderHint(QPainter::RenderHint::Antialiasing, true);

    painter.fillRect(getContentRect(), QColorConstants::White);
    QRectF contentRect = getContentRect();

    if (Engine::engine->baseTimelineItem() != nullptr)
    {
        drawBackgroundLines(&painter);

        // Draw line on left-hand side.
        // This separates the row labels from the content.
        int width = getContentRect().x() + getContentRect().width();
        int height = getContentRect().y() + getContentRect().height() * (videoToPlotRatio * analyzeAnimationAmount);

        // Draw line in center
        painter.setClipRect(getBelowPlayheadRect());
        QLineF centerLine(0, height, width, height);
        painter.setPen(QPen(QColor(0,0,0,90 * analyzeAnimationAmount), 2));
        painter.drawLine(centerLine);

        for (int i = 0; i < Engine::engine->timelineItems().size() + 1; i++)
        {
            painter.setClipRect(getBelowPlayheadRect());
            painter.setRenderHint(QPainter::RenderHint::Antialiasing, false);

            qreal lineY = getVideoRowRect(i).top();
            painter.setPen(QPen(QColor(0,0,0,40)));
            painter.drawLine(0, lineY, width, lineY);
            painter.setRenderHint(QPainter::RenderHint::Antialiasing, true);

            painter.setClipRect(getContentRect());
            drawVideoRow(&painter, i);
        }

        painter.setClipRect(getContentRect());

        painter.setClipRect(rect());
        drawPlotRow(&painter);

        drawScrollBar(&painter);
        drawPlayhead(&painter);
        drawCurrentValueDots(&painter);
    }
    else
    {
        QFont f(painter.font());
        f.setItalic(true);
        painter.setFont(f);
        painter.setPen(DefaultColorConstants::TextSecondary);
        QTextOption textOptions(Qt::AlignmentFlag::AlignCenter);
        painter.drawText(getContentRect(),
                         "No base video loaded",
                         textOptions);
    }

    // Gradient
    QLinearGradient g(getLeftLineOffset(), 0, getLeftLineOffset() + 10, 0);
    g.setColorAt(0, QColor(0, 0, 0, 25));
    g.setColorAt(1, QColor(0, 0, 0, 0));
    painter.fillRect(contentRect, g);

    auto pr = getPlayheadAreaRect();
    QLinearGradient g2(0, pr.y() + pr.height(), 0, pr.y() + pr.height() + 10);
    g2.setColorAt(0, QColor(0, 0, 0, 25));
    g2.setColorAt(1, QColor(0, 0, 0, 0));
    painter.fillRect(contentRect, g2);
}

void TimelineView::mousePressEvent(QMouseEvent *event)
{
    if (event->buttons() != Qt::MouseButton::LeftButton)
        return;

    // If there is no base timeline item loaded,
    // then the user should not be able to interact with
    // the timeline at all.
    if (Engine::engine->baseTimelineItem() == nullptr)
        return;

    m_previousMousePoint = event->localPos();
    m_mousePressPoint = event->x();

    if (getContentRect().contains(event->localPos()))
    {
        m_holding = HoldingItem::TimelineItem;
        mousePressEvent_ContentRect(event);
    }

    else if (getPlayheadAreaRect().contains(event->localPos()))
        m_holding = HoldingItem::Playhead;

    else if (getScrollBarLeftHandleRect(0).contains(event->localPos()))
        m_holding = HoldingItem::ScrollBarLeftHandle;

    else if (getScrollBarRightHandleRect(0).contains(event->localPos()))
        m_holding = HoldingItem::ScrollBarRightHandle;

    else if (getScrollBarCenterRect().contains(event->localPos()))
        m_holding = HoldingItem::ScrollBarCenter;
}

void TimelineView::mousePressEvent_ContentRect(QMouseEvent *event)
{
    if (Engine::engine->workspaceMode() == Engine::WorkspaceMode::Synchronize)
    {
        // Check if localPos is inside of any of the video rects.
        for (int i = 0; i < Engine::engine->timelineItems().size() + 1; i++)
        {
            auto thisItem = Engine::engine->timelineItem(i);
            if (thisItem == Engine::engine->baseTimelineItem()) continue;
            if (thisItem->locked()) continue;

            auto thisItemRect = getVideoItemRect(i);
            auto thisItemClicked = thisItemRect.contains(event->localPos());
            thisItem->setSelected(thisItemClicked);
        }

        // Check if localPos is inside the plot rect
        Engine::engine->setPlotSelected(!Engine::engine->plotLocked() && getPlotItemRect().contains(event->localPos()));
        emit selectionChanged();
    }

    update();
    event->accept();
}

void TimelineView::mouseMoveEvent(QMouseEvent *event)
{
    if (Engine::engine->baseTimelineItem() == nullptr)
        return;

    switch (m_holding)
    {
    case HoldingItem::Playhead:
        mouseMoveEvent_Playhead(event);
        break;
    case HoldingItem::TimelineItem:
        mouseMoveEvent_ContentRect(event);
        break;
    case HoldingItem::ScrollBarLeftHandle:
        mouseMoveEvent_ScrollBarHandle(event, &Engine::setLeftVisibleBound);
        break;
    case HoldingItem::ScrollBarRightHandle:
        mouseMoveEvent_ScrollBarHandle(event, &Engine::setRightVisibleBound);
        break;
    case HoldingItem::ScrollBarCenter:
        mouseMoveEvent_ScrollBarCenter(event);
        break;
    default:
        break;
    }

    m_previousMousePoint = event->localPos();
}

void TimelineView::mouseMoveEvent_Playhead(QMouseEvent *event)
{
    Engine::engine->setCurrentTime(rectXToTime(event->x()));
}

void TimelineView::mouseMoveEvent_ContentRect(QMouseEvent *event)
{
    auto tempTranslationLocal = event->x() - m_mousePressPoint;

    qreal inL = 0;
    qreal inR = rect().width() - getLeftLineOffset();

    qreal outL = Engine::engine->leftVisibleBound();
    qreal outR = Engine::engine->rightVisibleBound();

    auto tempTranslationTime = remap(inL, inR, outL, outR, tempTranslationLocal);

    auto atLeastOneClipSelected = false;

    if (Engine::engine->workspaceMode() == Engine::WorkspaceMode::Synchronize)
    {
        for (int i = 0; i < Engine::engine->timelineItems().size() + 1; i++)
        {
            auto thisItem = Engine::engine->timelineItem(i);
            if (!thisItem->selected()) continue;
            if (thisItem == Engine::engine->baseTimelineItem()) continue;
            if (thisItem->locked()) continue;

            atLeastOneClipSelected = true;
            thisItem->setTemporaryTranslation(tempTranslationTime);
        }

        if (!Engine::engine->plotLocked() && Engine::engine->plotSelected())
        {
            atLeastOneClipSelected = true;
            Engine::engine->setPlotTemporaryTranslation(tempTranslationTime);
        }

    }

    // No clips were selected, so move the playhead to this position
    if (!atLeastOneClipSelected)
    {
        Engine::engine->setCurrentTime(rectXToTime(event->x()));
    }
    else
    {
        update();
    }

    event->accept();
}

void TimelineView::mouseMoveEvent_ScrollBarHandle(QMouseEvent *event, void (Engine::*setBoundFunction)(qreal))
{
    qreal x = event->localPos().x();

    qreal inL = getScrollBarContainerRect().left();
    qreal inR = getScrollBarContainerRect().right();

    qreal outL = 0;
    qreal outR = Engine::engine->durationSecs();

    (Engine::engine->*setBoundFunction)(remap(inL, inR, outL, outR, x));
    event->accept();
}

void TimelineView::mouseMoveEvent_ScrollBarCenter(QMouseEvent *event)
{
    // Get offset of cursor relative to left side of scroll bar
    qreal offset = event->x() - getScrollBarRect().left();

    // Convert offset to seconds
    qreal offsetInL = 0;
    qreal offsetInR = getScrollBarContainerRect().width();
    qreal offsetOutL = 0;
    qreal offsetOutR = Engine::engine->durationSecs();
    qreal offsetTime = remap(offsetInL, offsetInR, offsetOutL, offsetOutR, offset);

    // Convert event x to time
    qreal posInL = getScrollBarContainerRect().left();
    qreal posInR = getScrollBarContainerRect().right();
    qreal posOutL = 0;
    qreal posOutR = Engine::engine->durationSecs();
    qreal posTime = remap(posInL, posInR, posOutL, posOutR, event->x());

    qreal range = Engine::engine->rightVisibleBound() - Engine::engine->leftVisibleBound();

    qreal xDelta = event->x() - m_previousMousePoint.x();
    qreal tDelta = remap(offsetInL, offsetInR, offsetOutL, offsetOutR, xDelta);

    // Calculate what the new left and right bounds would be
    qreal newLeftBound = posTime + tDelta - offsetTime;
    qreal newRightBound = newLeftBound + range;

    if (tDelta > 0 && newRightBound > Engine::engine->durationSecs())
    {
        // Right bound would go past duration!
        newLeftBound = Engine::engine->durationSecs() - range;
        newRightBound = Engine::engine->durationSecs();
    }
    else if (tDelta < 0 && newLeftBound < 0)
    {
        // Left bound would go past zero!
        newLeftBound = 0;
        newRightBound = range;
    }

    Engine::engine->setLeftVisibleBound(newLeftBound);
    Engine::engine->setRightVisibleBound(newRightBound);
    event->accept();
}

void TimelineView::mouseReleaseEvent(QMouseEvent *event)
{
    if (Engine::engine->baseTimelineItem() == nullptr)
        return;

    if (Engine::engine->workspaceMode() == Engine::WorkspaceMode::Synchronize)
    {
        for (int i = 0; i < Engine::engine->timelineItems().size() + 1; i++)
        {
            auto thisItem = Engine::engine->timelineItem(i);
            auto isTimelineItem = thisItem == Engine::engine->baseTimelineItem();

            if (getVideoVisibilityButtonRect(i).contains(event->localPos()))
            {
                thisItem->setVisible(!thisItem->visible());
            }

            else if (getVideoLockButtonRect(i).contains(event->localPos()) && !isTimelineItem)
            {
                thisItem->setLocked(!thisItem->locked());
            }

            else if (!thisItem->locked() && !isTimelineItem)
            {
                thisItem->commitTemporaryTranslation();
            }
        }

        // Check if we want to lock/unlock plot
        if (getPlotLockButtonRect().contains(event->localPos()))
        {
            Engine::engine->setPlotLocked(!Engine::engine->plotLocked());
        }

        // Check if we want to commit plot translation...?
        if (getPlotItemRect().contains(event->localPos()))
        {
            Engine::engine->commitPlotTemporaryTranslation();
        }
    }

    update();
    event->accept();

    m_holding = HoldingItem::None;
}

void TimelineView::onContextMenuRequested(const QPoint &point)
{
    if (Engine::engine->baseTimelineItem() == nullptr)
        return;

    if (Engine::engine->workspaceMode() != Engine::WorkspaceMode::Synchronize)
        return;

    // Create a context menu with an item to delete the selected video row.
    QMenu ctx(this);

    // Name of the item (not clickable)
    QAction currentNameAction(this);
    currentNameAction.setEnabled(false);

    QAction deleteRowAction("Delete", this);

    QAction renameAction("Rename...", this);



    QAction setOffsetAction("Set Offset...", this);

    QAction lockAction("Locked", this);
    lockAction.setCheckable(true);

    QAction showAction("Visible", this);
    showAction.setCheckable(true);

    ctx.addAction(&currentNameAction);
    ctx.addSeparator();
    ctx.addAction(&deleteRowAction);
    ctx.addAction(&renameAction);
    ctx.addAction(&setOffsetAction);
    ctx.addSeparator();
    ctx.addAction(&lockAction);
    ctx.addAction(&showAction);

    // Clicked on plot row!
    if (getPlotRowRect().contains(QPointF(point)))
    {
        auto currentPlot = Engine::engine->currentPlot();
        if (currentPlot == nullptr)
        {
            currentNameAction.setText("No plot loaded");
            setOffsetAction.setEnabled(false);
            lockAction.setEnabled(false);
        }
        else
        {
            currentNameAction.setText(currentPlot->name());
            showAction.setChecked(true);
            lockAction.setChecked(Engine::engine->plotLocked());
            connect(&lockAction, &QAction::triggered, Engine::engine, &Engine::setPlotLocked);

            connect(&setOffsetAction, &QAction::triggered,
                    [=]()
                    {
                        bool ok = false;
                        auto newOffset = QInputDialog::getDouble(
                                    this, "Set Offset", "Offset:",
                                    Engine::engine->plotOffset(), -2147483647, 2147483647, 3, &ok);
                        if (ok)
                        {
                            Engine::engine->setPlotOffset(newOffset);
                        }
                    });
        }

        deleteRowAction.setEnabled(false);
        renameAction.setEnabled(false);
        showAction.setEnabled(false);

        ctx.exec(mapToGlobal(point));
        return;
    }


    // Determine if we're in a video row
    int rowNum = -1;
    TimelineItem *item = nullptr;

    for (int i = 0; i < Engine::engine->timelineItems().count() + 1; i++)
    {
        auto row = getVideoRowRect(i);
        if (row.contains(QPointF(point)))
        {
            rowNum = i;
            item = Engine::engine->timelineItem(i);
            break;
        }
    }

    // Clicked on a video row!
    if (rowNum == -1)
        return;

    currentNameAction.setText(item->name());

    // Action for deleting video (only enabled if it's not the base video)

    connect(&deleteRowAction, &QAction::triggered, [=]() { Engine::engine->deleteTimelineItem(item);});
    deleteRowAction.setEnabled(rowNum > 0);

    // Action for renaming video
    connect(&renameAction, &QAction::triggered,
            [=]()
            {
                bool ok = false;
                auto newName = QInputDialog::getText(
                            this, "Rename Video", "New name:",
                            QLineEdit::Normal, item->name(), &ok);
                if (ok)
                {
                    item->setName(newName);
                }
            });

    // Action for changing offset of item
    setOffsetAction.setEnabled(!item->isBaseLayer());
    connect(&setOffsetAction, &QAction::triggered,
            [=]()
            {
                bool ok = false;
                auto newOffset = QInputDialog::getDouble(
                            this, "Set Offset", "Offset:",
                            item->offset(), -2147483647, 2147483647, 3, &ok);
                if (ok)
                {
                    item->setOffset(newOffset);
                }
            });

    // Actions for locking/unlocking, and showing/hiding
    lockAction.setChecked(item->locked());
    lockAction.setEnabled(rowNum > 0);
    connect(&lockAction, &QAction::triggered, item, &TimelineItem::setLocked);

    showAction.setChecked(item->visible());
    connect(&showAction, &QAction::triggered, item, &TimelineItem::setVisible);

    ctx.exec(mapToGlobal(point));
}

qreal TimelineView::getLeftLineOffset()
{
    return leftLineOffsetAnalyzeMode;
}

QRectF TimelineView::getBelowPlayheadRect()
{
    QRectF labelsRect(rect());
    labelsRect.setTop(playheadAreaHeight);
    labelsRect.setBottom(labelsRect.y() + labelsRect.height() - scrollBarHeight);
    return labelsRect;
}

QRectF TimelineView::getScrollBarContainerRect()
{
    QRectF containerRect(rect());
    containerRect.setTop(containerRect.y() + containerRect.height() - scrollBarHeight);
    containerRect.setLeft(getLeftLineOffset());
    return containerRect;
}

QRectF TimelineView::getLabelsRect()
{
    QRectF labelsRect(getBelowPlayheadRect());
    labelsRect.setRight(getLeftLineOffset() - 0.5);
    return labelsRect;
}

QRectF TimelineView::getPlayheadAreaRect()
{
    return QRectF(getLeftLineOffset(), 0, rect().width() - getLeftLineOffset(), playheadAreaHeight);
}

/*!
 * \brief The rectangle with the left edge at the border of the timeline
 * and the bottom edge at above the scroll bar.
 *
 * The sides of this rectangle should be at the following positions:
 * \list
 *  \li \c top() - top edge of the TimelineView
 *  \li \c bottom() - top edge of the scroll bar
 *  \li \c left() - edge where the vertical axis tick marks are
 *  \li \c right() - right edge of the TimelineView
 * \endlist
 */
QRectF TimelineView::getTimelineRect()
{
    QRectF timelineRect = rect();
    timelineRect.setLeft(getLeftLineOffset());
    timelineRect.setBottom(timelineRect.y() + timelineRect.height() - scrollBarHeight);
    return timelineRect;
}


/*!
 * \brief The rectangle containing all of the video and plot items.
 *
 * The sides of this rectangle should be at the following positions:
 * \list
 *  \li \c top() - edge where the horizontal axis tick marks are
 *  \li \c bottom() - top edge of the scroll bar
 *  \li \c left() - edge where the vertical axis tick marks are
 *  \li \c right() - right edge of the TimelineView
 * \endlist
 */
QRectF TimelineView::getContentRect()
{
    QRectF contentRect(getBelowPlayheadRect());
    contentRect.setLeft(getLeftLineOffset());
    return contentRect;
}


/*!
 * \brief The rectangle for the row that the video item at \a index occupies.
 *
 * The sides of this rectangle should be at the following positions:
 * \list
 *  \li \c top() - top edge of the given video item's row
 *  \li \c bottom() - bottom edge of the given video item's row
 *  \li \c left() - left edge of the TimelineView
 *  \li \c right() - right edge of the TimelineView
 * \endlist
 */
QRectF TimelineView::getVideoRowRect(int index)
{
    // Get the rect representing the row
    QRectF rowRect(0,
                   getContentRect().y() + getContentRect().height() * (videoToPlotRatio * analyzeAnimationAmount) - videoRowHeight,
                   rect().width(), videoRowHeight);
    rowRect.translate(0, -videoRowHeight * index);
    return rowRect;
}


/*!
 * \brief The rectangle for the video item at \a index.
 *
 * The sides of this rectangle should be at the following positions:
 * \list
 * \li \c top() - top edge of the given video item's row
 * \li \c bottom() - bottom edge of the given video item's row
 * \li \c left() - left edge of the video item's extent
 * \li \c right() - right edge of the video item's extent
 * \endlist
 */
QRectF TimelineView::getVideoItemRect(int index)
{
    auto timelineItem = Engine::engine->timelineItem(index);

    if (timelineItem == nullptr)
    {
        return QRectF();
    }
    auto start = timelineItem->displayOffset();
    auto duration = timelineItem->duration();
    auto end = start + duration;

    qreal remappedStart = timeToRectX(start);
    qreal remappedEnd = timeToRectX(end);

    QRectF videoItemRect = getVideoRowRect(index);
    videoItemRect.setLeft(remappedStart);
    videoItemRect.setRight(remappedEnd);

    return videoItemRect;
}


/*!
 * \brief The rectangle for the Lock/Unlock button on the video item row
 * at \a index.
 */
QRectF TimelineView::getVideoLockButtonRect(int index)
{
    QRectF lockIconRect(getVideoRowRect(index));
    lockIconRect.setHeight(lockIconRect.height() * 0.75);
    lockIconRect.setWidth(lockIconRect.height());

    lockIconRect.translate(0, videoRowHeight / 2 - lockIconRect.height() / 2);
    return lockIconRect;
}


/*!
 * \brief The rectangle for the Show/Hide button on the video item row
 * at \a index.
 */
QRectF TimelineView::getVideoVisibilityButtonRect(int index)
{
    QRectF visibilityIconRect(getVideoRowRect(index));
    visibilityIconRect.setHeight(visibilityIconRect.height() * 0.75);
    visibilityIconRect.setWidth(visibilityIconRect.height());

    visibilityIconRect.translate(visibilityIconRect.width() + timelineButtonSpacing, videoRowHeight / 2 - visibilityIconRect.height() / 2);
    return visibilityIconRect;
}


/*!
 * \brief The rectangle for the Lock/Unlock button on the plot item row.
 */
QRectF TimelineView::getPlotLockButtonRect()
{
    QRectF lockIconRect(getPlotRowRect());
    lockIconRect.setHeight(videoRowHeight * 0.75);
    lockIconRect.setWidth(lockIconRect.height());
    lockIconRect.translate(0, getPlotRowRect().height() / 2 - lockIconRect.height() / 2);
    return lockIconRect;
}

QRectF TimelineView::getPlotVisibilityButtonRect()
{
    QRectF lockIconRect(getPlotRowRect());
    lockIconRect.setHeight(videoRowHeight * 0.75);
    lockIconRect.setWidth(lockIconRect.height());
    lockIconRect.translate(lockIconRect.width() + timelineButtonSpacing, getPlotRowRect().height() / 2 - lockIconRect.height() / 2);
    return lockIconRect;
}

/*!
 * \brief The rectangle for the row that the plot occupies.
 */
QRectF TimelineView::getPlotRowRect()
{
    QRectF plotRowRect(0, getContentRect().y() + getContentRect().height() * (videoToPlotRatio * analyzeAnimationAmount),
                       rect().width(), getContentRect().height() * (1.0f - videoToPlotRatio * analyzeAnimationAmount)
                       );
    return plotRowRect;
}

QRectF TimelineView::getPlotItemRect()
{
    QRectF plotRect = getPlotRowRect();
    plotRect.setLeft(timeToRectX(Engine::engine->plotDisplayOffset()));

    if (Engine::engine->currentPlot() == nullptr)
    {
        plotRect.setRight(timeToRectX(Engine::engine->plotDisplayOffset()));
    }
    else
    {
        plotRect.setRight(timeToRectX(Engine::engine->plotDisplayOffset() + Engine::engine->currentPlot()->duration()));
    }
    return plotRect;
}

void TimelineView::drawTopPart(QPainter *painter)
{
    QLineF line(0, playheadAreaHeight, rect().width(), playheadAreaHeight);
    painter->setPen(DefaultColorConstants::LinePrimary);
    painter->drawLine(line);
}

void TimelineView::drawBackgroundLines(QPainter *painter)
{
    QRectF contentRect = getContentRect();

    qreal leftBound = Engine::engine->leftVisibleBound();
    qreal rightBound = Engine::engine->rightVisibleBound();

    qreal range = rightBound - leftBound;

    qreal intervalsPerPixel = range / getContentRect().width();

    qreal scale = intervalsPerPixel;
    qreal cyPerInch = 20.;

    qreal xUnitsPerGrid;
    int xDecPlaces;
    std::tie(xUnitsPerGrid, xDecPlaces) = PlotMath::computeUnitsPerGrid(cyPerInch, scale);

    qreal tickLocation = 0;

    if (leftBound >= 0)
        while (tickLocation < leftBound) tickLocation += xUnitsPerGrid;
    else
        while (tickLocation > leftBound) tickLocation -= xUnitsPerGrid;

    qreal lineX = getContentRect().left();


    QString labelText;
    while (lineX < getContentRect().right())
    {

        lineX = timeToRectX(tickLocation);

        labelText = QString::number(tickLocation);

        tickLocation += xUnitsPerGrid;

        QLineF line(lineX, contentRect.y() - 5, lineX, contentRect.y() + contentRect.height());

        painter->setPen(DefaultColorConstants::LineSecondary);
        painter->drawLine(line);

        painter->setPen(QColor(0,0,0));

        qreal textWidth = painter->fontMetrics().boundingRect(labelText).width();
        painter->drawText(lineX - textWidth / 2, contentRect.top() - 10, labelText);
    }
}

QRectF TimelineView::getScrollBarRect()
{
    QRectF bar = getScrollBarContainerRect();
    qreal durationReal = Engine::engine->durationSecs();
    qreal left = getScrollBarContainerRect().left();
    qreal width = getScrollBarContainerRect().width();

    bar.setLeft(left + width * (Engine::engine->leftVisibleBound() / durationReal));
    bar.setRight(left + width * (Engine::engine->rightVisibleBound() / durationReal));

    return bar;
}

QRectF TimelineView::getScrollBarLeftHandleRect(qreal handlePadding)
{
    QRectF leftHandle(getScrollBarRect());

    leftHandle.setRight(leftHandle.x() + scrollBarRadius * 2 - handlePadding);
    leftHandle.setLeft(leftHandle.x() + handlePadding);
    leftHandle.setBottom(leftHandle.y() + leftHandle.height() - handlePadding);
    leftHandle.setTop(leftHandle.y() + handlePadding);
    return leftHandle;
}

QRectF TimelineView::getScrollBarRightHandleRect(qreal handlePadding)
{
    QRectF rightHandle(getScrollBarRect());
    rightHandle.setLeft(rightHandle.x() + rightHandle.width() - scrollBarRadius * 2 + handlePadding);
    rightHandle.setRight(rightHandle.x() + rightHandle.width() - handlePadding);
    rightHandle.setBottom(rightHandle.y() + rightHandle.height() - handlePadding);
    rightHandle.setTop(rightHandle.y() + handlePadding);
    return rightHandle;
}

QRectF TimelineView::getScrollBarCenterRect()
{
    QRectF bar = getScrollBarRect();
    bar.setLeft(getScrollBarLeftHandleRect(0).right());
    bar.setRight(getScrollBarRightHandleRect(0).left());
    return bar;
}

void TimelineView::drawScrollBar(QPainter *painter)
{
    painter->setClipRect(getScrollBarContainerRect());

    painter->setPen(Qt::NoPen);
    painter->setBrush(DefaultColorConstants::ScrollBarBackground);
    painter->drawRoundedRect(getScrollBarContainerRect(), scrollBarRadius, scrollBarRadius);

    // Draw the bar itself
    QRectF bar = getScrollBarRect();

    // Draw the handles on the sides of the bar
    QRectF leftHandle = getScrollBarLeftHandleRect(scrollBarHandlePadding);

    // Right handle
    QRectF rightHandle = getScrollBarRightHandleRect(scrollBarHandlePadding);

    painter->setPen(DefaultColorConstants::ScrollBarBodyOutline);
    painter->setBrush(DefaultColorConstants::ScrollBarBody);
    painter->drawRoundedRect(bar, scrollBarRadius, scrollBarRadius);

    // Grip lines in center
    qreal centerLinePadding = 4;
    qreal cx = getScrollBarCenterRect().center().x();
    QLineF centerLine(
                cx,
                getScrollBarCenterRect().y() + centerLinePadding,
                cx,
                getScrollBarCenterRect().y() + getScrollBarCenterRect().height() - centerLinePadding
                );

    QLineF leftLine(centerLine);
    leftLine.translate(-centerLinePadding, 0);
    QLineF rightLine(centerLine);
    rightLine.translate(centerLinePadding, 0);

    painter->setPen(DefaultColorConstants::ScrollBarVerticalLines);
    painter->drawLine(centerLine);
    painter->drawLine(leftLine);
    painter->drawLine(rightLine);

    painter->setPen(Qt::NoPen);
    painter->setBrush(DefaultColorConstants::ScrollBarHandle);
    painter->drawEllipse(leftHandle);
    painter->drawEllipse(rightHandle);
}

void TimelineView::drawVideoRow(QPainter *painter, int index)
{
    // Draw video rows, but clipped to their area
    QRectF videoRowRect = getVideoRowRect(index);
    QRectF videoItemRect = getVideoItemRect(index);

    auto timelineItem = Engine::engine->timelineItem(index);
    auto isBaseItem = timelineItem == Engine::engine->baseTimelineItem();
    auto selected = timelineItem->selected();

    // Highlight row backgrounds if they are currently selected
    painter->setPen(Qt::NoPen);
    painter->setBrush(QColor(0, 0, 0, 20));
    if (selected) painter->drawRect(videoRowRect);


    // Draw the video item itself.
    // The base video has a different color than the additional videos.
    QColor color = isBaseItem ? DefaultColorConstants::BaseVideo : DefaultColorConstants::AdditionalVideo;
    color.setAlphaF(color.alphaF() * analyzeAnimationAmount);

    // If a video item is selected, it should have a lighter color.
    painter->setBrush(QBrush(color.lighter(selected ? 170 : 150)));
    painter->drawRect(videoItemRect);


    // Draw border for the row.
    painter->setRenderHint(QPainter::RenderHint::Antialiasing, false);
    painter->setPen(QPen(color, 1));
    painter->setBrush(Qt::NoBrush);
    painter->drawRect(videoItemRect);
    painter->setRenderHint(QPainter::RenderHint::Antialiasing, true);


    // Draw the text indicating what the video item's offset is.
    QFont offsetTextFont;
    offsetTextFont.setItalic(false);
    painter->setFont(offsetTextFont);

    QColor whiteColor(255,255,255,255 * analyzeAnimationAmount);
    QTextOption offsetTextOptions(Qt::AlignmentFlag::AlignVCenter | Qt::AlignmentFlag::AlignLeft);
    offsetTextOptions.setWrapMode(QTextOption::NoWrap);
    painter->setPen(whiteColor);

    QRectF offsetTextRect = videoItemRect;
    offsetTextRect.setLeft((offsetTextRect & getContentRect()).left() + 5.0);
    painter->drawText(offsetTextRect,
                      QString("%1 (%2%3s)")
                      .arg(timelineItem->name(),
                           timelineItem->displayOffset() >= 0 ? "+" : "",
                           QString::number(timelineItem->displayOffset(), 'g', 3)),
                      offsetTextOptions);

    // Draw buttons on left
    QRectF lockButtonRect = getVideoLockButtonRect(index);
    qreal lockButtonOpacity = (timelineItem->isBaseLayer() ? 0.25 : 1.0) * analyzeAnimationAmount;
    QColor iconColor(QColorConstants::Black);

    painter->setClipRect(getBelowPlayheadRect());

    if (timelineItem->locked())
        m_lockFillRenderer.render(painter, iconColor, lockButtonRect, lockButtonOpacity);
    else
        m_lockDisabledRenderer.render(painter, iconColor, lockButtonRect, lockButtonOpacity);

    QRectF visibilityButtonRect = getVideoVisibilityButtonRect(index);
    if (timelineItem->visible())
        m_visibilityFillRenderer.render(painter, iconColor, visibilityButtonRect, analyzeAnimationAmount);
    else
        m_visibilityDisabledRenderer.render(painter, iconColor, visibilityButtonRect, analyzeAnimationAmount);
}

void TimelineView::drawPlotRow(QPainter *painter)
{
    painter->setClipRect(getBelowPlayheadRect());

    bool selected = Engine::engine->plotSelected();
    // highlight the row
    painter->setPen(Qt::NoPen);
    painter->setBrush(QColor(0, 0, 0, 20 * (1.0f - analyzeAnimationAmount)));
    if (selected) painter->drawRect(getPlotRowRect());

    painter->setBrush(Qt::NoBrush);

    QRectF plotRowRect = getPlotRowRect();

    QRectF plotRowLabelRect = plotRowRect & getLabelsRect();
    plotRowLabelRect.setRight(getLeftLineOffset() - 5);

    // Draw lock button
    (Engine::engine->plotLocked() ? m_lockFillRenderer : m_lockDisabledRenderer)
            .render(painter, QColor(0,0,0), getPlotLockButtonRect(), analyzeAnimationAmount);

    m_visibilityFillRenderer.render(painter, QColorConstants::Black, getPlotVisibilityButtonRect(), 0.25 * analyzeAnimationAmount);

    // Draw tick marks if we're in Analyze mode
    painter->setClipRect(getPlotRowRect());
    drawVerticalTickMarks(painter);


    // Draw the vertical bar colored with the quantities that are in the same
    // scale group as the current Y axis quantity.
    drawVerticalQuantityColorBar(painter);

    QRectF plotItemRect = getPlotItemRect();
    QRectF offsetTextRect = plotItemRect & getContentRect();
    offsetTextRect.setBottom(plotItemRect.top() + videoRowHeight * analyzeAnimationAmount);


    // Draw stuff
    painter->setClipRect(plotRowRect & getContentRect());

    QColor color = DefaultColorConstants::PlotItem;
    color.setAlphaF(color.alphaF() * analyzeAnimationAmount);
    painter->setPen(QPen(color, 2));
    painter->setBrush(Qt::NoBrush);
    painter->drawRect(plotItemRect);

    // Background fill should only be at the top
    painter->fillRect(offsetTextRect, QBrush(color.lighter(selected ? 140 : 120)));

    QRectF plotItemPreviewRect = plotItemRect;
    plotItemPreviewRect.setTop(plotItemPreviewRect.top() + videoRowHeight * analyzeAnimationAmount);
    painter->fillRect(plotItemPreviewRect, QColor(255, 255, 255, 255 * analyzeAnimationAmount));
    if (m_plotImage != nullptr)
    {
        painter->drawImage(plotItemPreviewRect, *m_plotImage);
    }

    // Draw text at top
    if (Engine::engine->currentPlot() != nullptr)
    {
        QTextOption offsetTextOptions(Qt::AlignmentFlag::AlignVCenter | Qt::AlignmentFlag::AlignLeft);
        offsetTextOptions.setWrapMode(QTextOption::NoWrap);
        painter->setPen(QColor(255,255,255,255 * analyzeAnimationAmount));
        offsetTextRect.setLeft(offsetTextRect.left() + 5.0);
        painter->drawText(offsetTextRect,
                          QString("%1 (%2%3s)")
                          .arg(Engine::engine->currentPlot()->name(),
                               Engine::engine->plotDisplayOffset() >= 0 ? "+" : "",
                               QString::number(Engine::engine->plotDisplayOffset(), 'g', 3)),
                          offsetTextOptions);
    }
    else
    {
        QFont font = painter->font();
        font.setItalic(true);
        painter->setFont(font);
        painter->setPen(DefaultColorConstants::TextSecondary);

        painter->drawText(getPlotRowRect() & getContentRect(),
                          "No plot loaded",
                          QTextOption(Qt::AlignmentFlag::AlignCenter));
    }

    painter->setClipRect(rect());
}

void TimelineView::drawVerticalQuantityColorBar(QPainter *painter)
{
    qreal barWidth = QUANTITY_BAR_THICKNESS;
    auto yAxisQuantities = Engine::engine->quantitiesForCurrentScaleGroup();
    auto singleRectHeight = (getPlotRowRect().height() - 2 * barWidth) / yAxisQuantities.length();
    QRectF singleColorRect = QRect(0, getPlotRowRect().top(), barWidth, singleRectHeight);
    painter->setPen(Qt::NoPen);
    for (int i = 0; i < yAxisQuantities.length(); i++)
    {
        auto thisRect = singleColorRect.translated(0, barWidth + singleRectHeight * i);

        // Make color
        auto quantityColor = yAxisQuantities.at(i)->color();
        quantityColor.setAlphaF(1 - analyzeAnimationAmount);
        painter->setBrush(quantityColor);
        painter->drawRect(thisRect);
        auto r = thisRect.width() / 2;

        // First rectangle should have a round cap on top
        if (i == 0)
        {
            auto topCenter = QPointF(thisRect.center().x(), thisRect.top());
            painter->drawEllipse(topCenter, r, r);
        }

        // Last rectangle should have a round cap on bottom
        // (Note: this can't be an else if statement, because it's possible
        // that there will be only one rectangle, and thus that there should
        // be both caps.)
        if (i == yAxisQuantities.length() - 1)
        {
            auto bottomCenter = QPointF(thisRect.center().x(), thisRect.y() + thisRect.height());
            painter->drawEllipse(bottomCenter, r, r);
        }
    }
}

void TimelineView::drawVerticalTickMarks(QPainter *painter)
{
    qreal cyPerInch = 5.;
    auto yQuant = Engine::engine->plotYQuantity();
    if (yQuant == nullptr)
        return;

    qreal physicalSize = getPlotRowRect().height();

    TickMarkData data = PlotMath::calculateTickMarks(
                Engine::engine->plotYQuantity(), true,
                cyPerInch, physicalSize);

    if (!data.validData)
    {
        return;
    }

    qreal tickLocation = data.startingTickLocation;
    qreal lineY = 0;
    while (lineY < 1)
    {
        lineY = invLerp(data.bottomBound, data.topBound, tickLocation);
        drawVerticalTickMark(painter, lineY, tickLocation);
        tickLocation += data.unitsPerGrid;
    }
}


/*
 * This function takes in three parameters:
 * - painter: a pointer to the QPainter
 * - t: a number that determines where to draw the tick mark. 0 corresponds to
 *      the bottom of the plot area, and 1 corresponds to the top. This value
 *      can be less than 0 or greater than 1
 * - value: the number to display next to the tick mark. This has no bearing
 *          on where the tick mark itself is painted (that comes from t)
 */
void TimelineView::drawVerticalTickMark(QPainter *painter, qreal t, qreal value)
{
    QRectF r = getPlotRowRect() & getContentRect();
    qint32 leftYWidgetSpace = lerp(r.y() + r.height(), r.y(), t);
    QPoint p1(getContentRect().left(), leftYWidgetSpace);
    QPoint p2(getContentRect().left() - TICK_LENGTH, leftYWidgetSpace);
    QPoint p3(getContentRect().right(), leftYWidgetSpace);

    auto col = DefaultColorConstants::LineSecondary;
    col.setAlphaF(1 - analyzeAnimationAmount);
    painter->setPen(col);
    painter->drawLine(p2, p3);

    col = QColorConstants::Gray;
    col.setAlphaF(1 - analyzeAnimationAmount);
    painter->setPen(col);
    painter->drawLine(p1, p2);

    col = DefaultColorConstants::Text;
    col.setAlphaF(1 - analyzeAnimationAmount);
    painter->setPen(col);

    QString labelText = QString::number(value);
    QRect textRect = painter->fontMetrics().boundingRect(labelText);
    painter->drawText(p2.x() - textRect.width() - TICK_PADDING, p2.y() + textRect.height() / 2, labelText);
}

void TimelineView::drawPlayhead(QPainter *painter)
{
    qreal pos = timeToRectX(Engine::engine->currentTime());


    auto triangle = getPlayheadTriangle();
    triangle.translate(pos, playheadAreaHeight);

    auto trianglePoint = triangle[0];
    auto bottomPoint = QPointF(trianglePoint);
    bottomPoint.setY(rect().height());

    painter->setClipRect(getTimelineRect());
    painter->setPen(QColor(0,0,0).lighter());
    painter->drawLine(trianglePoint, bottomPoint);

    painter->setPen(QColor(255,0,0).darker());
    painter->setBrush(QColor(255,0,0));
    painter->drawPolygon(triangle);
}

void TimelineView::drawCurrentValueDots(QPainter *painter)
{

    qreal plotTime = Engine::engine->currentTime() - Engine::engine->plotDisplayOffset();
    qreal xPos = timeToRectX(Engine::engine->currentTime());

    auto plot = Engine::engine->currentPlot();

    if (plot == nullptr)
        return;

    for (auto quantity : plot->quantities())
    {
        qreal yPos = plot->relativeValueForQuantityAtTime(quantity, plotTime);

        qreal top = getPlotRowRect().y();
        qreal bottom = getPlotRowRect().y() + getPlotRowRect().height();
        if (!plot->scaleGroups().value(quantity->scaleGroup())->railToRail())
        {
            qreal padding = (top - bottom) * (1 - RAIL_TO_RAIL_RATIO);
            top -= padding;
            bottom += padding;
        }

        painter->setBrush(quantity->color());
        painter->setPen(QPen(QColor(quantity->color()).darker(140), 1));

        qreal dotRadius = DOT_RADIUS * (1 - analyzeAnimationAmount);
        painter->drawEllipse(xPos - dotRadius,
                             lerp(bottom, top, yPos) - dotRadius,
                             dotRadius * 2, dotRadius * 2);
    }
}

void TimelineView::setSynchronizeMode()
{
    Engine::engine->setWorkspaceMode(Engine::WorkspaceMode::Synchronize);

    QVariantAnimation* anim = new QVariantAnimation(this);
    anim->setStartValue(0.0f);
    anim->setEndValue(1.0f);
    anim->setDuration(ANIM_TIME);
    anim->setEasingCurve(QEasingCurve::OutCubic);
    connect(anim, &QVariantAnimation::valueChanged,
            [=](QVariant value)
    {
        analyzeAnimationAmount = value.toFloat();
        update();
    }
    );
    connect(anim, &QAbstractAnimation::finished,
            [=]()
    {
        anim->deleteLater();
        invalidatePlot();
    }
    );
    anim->start();
}

void TimelineView::setAnalyzeMode()
{
    Engine::engine->setWorkspaceMode(Engine::WorkspaceMode::Analyze);
    Engine::engine->setPlotSelected(false);

    QVariantAnimation* anim = new QVariantAnimation(this);
    anim->setStartValue(1.0f);
    anim->setEndValue(0.0f);
    anim->setDuration(ANIM_TIME);
    anim->setEasingCurve(QEasingCurve::OutCubic);
    connect(anim, &QVariantAnimation::valueChanged,
            [=](QVariant value)
    {
        analyzeAnimationAmount = value.toFloat();
        update();
    }
    );
    connect(anim, &QAbstractAnimation::finished,
            [=]()
    {
        anim->deleteLater();
        invalidatePlot();
    }
    );
    anim->start();
}

// Math converters
qreal TimelineView::timeToRectX(qreal time)
{
    qreal timeL = Engine::engine->leftVisibleBound();
    qreal timeR = Engine::engine->rightVisibleBound();

    qreal rectL = getLeftLineOffset();
    qreal rectR = rect().width();

    return remap(timeL, timeR, rectL, rectR, time);
}

qreal TimelineView::rectXToTime(qreal x)
{
    qreal timeL = Engine::engine->leftVisibleBound();
    qreal timeR = Engine::engine->rightVisibleBound();

    qreal rectL = getLeftLineOffset();
    qreal rectR = rect().width();

    return remap(rectL, rectR, timeL, timeR, x);
}

// Graphics generators
QPolygonF TimelineView::getPlayheadTriangle()
{
    qreal xSize = PLAYHEAD_WIDTH / 2.;
    qreal ySize = PLAYHEAD_HEIGHT / 2.;
    QPolygonF p;
    p << QPointF(0,0) << QPointF(-xSize, -ySize) << QPointF(xSize, -ySize);
    return p;
}
