#ifndef TIMELINEVIEW_H
#define TIMELINEVIEW_H

#include <QPainter>
#include <QWidget>
#include <QPaintEvent>
#include <QMouseEvent>
#include <QDebug>
#include <QMenu>
#include <QInputDialog>
#include <QEvent>
#include <QHelpEvent>
#include <QToolTip>

#include "Models/Engine/engine.h"

#include "Utilities/graphicsmath.h"
#include "Utilities/SVGColorRenderer/svgcolorrenderer.h"
#include "Utilities/plotmath.h"
#include "Utilities/PlotDrawer/plotdrawer.h"
#include "Utilities/Constants/colorconstants.h"
#include "Utilities/Constants/drawingconstants.h"


class TimelineView : public QWidget
{
    Q_OBJECT
public:
    explicit TimelineView(QWidget *parent = nullptr);

protected:
    QSize sizeHint() const override;

    bool event(QEvent* event) override;

    void toolTipEvent(QHelpEvent* event);
    void paintEvent(QPaintEvent *event) override;
    void resizeEvent(QResizeEvent* event) override;

    void mousePressEvent(QMouseEvent *event) override;
    void mousePressEvent_ContentRect(QMouseEvent *event);

    void mouseMoveEvent(QMouseEvent *event) override;
    void mouseMoveEvent_Playhead(QMouseEvent *event);
    void mouseMoveEvent_ContentRect(QMouseEvent *event);
    void mouseMoveEvent_ScrollBarHandle(QMouseEvent *event, void (Engine::*setBoundFunction)(qreal));
    void mouseMoveEvent_ScrollBarCenter(QMouseEvent *event);

    void mouseReleaseEvent(QMouseEvent *event) override;

    qreal getLeftLineOffset();
    QRectF getBelowPlayheadRect();

    QRectF getScrollBarContainerRect();
    QRectF getScrollBarRect();
    QRectF getScrollBarLeftHandleRect(qreal handlePadding);
    QRectF getScrollBarRightHandleRect(qreal handlePadding);
    QRectF getScrollBarCenterRect();

    QRectF getLabelsRect();
    QRectF getTimelineRect();
    QRectF getContentRect();
    QRectF getVideoRowRect(int index);
    QRectF getVideoItemRect(int index);
    QRectF getPlotRowRect();
    QRectF getPlotItemRect();
    QRectF getVideoLockButtonRect(int index);
    QRectF getVideoVisibilityButtonRect(int index);
    QRectF getPlotLockButtonRect();
    QRectF getPlotVisibilityButtonRect();
    QRectF getPlayheadAreaRect();

    void drawTopPart(QPainter *painter);
    void drawBackgroundLines(QPainter *painter);
    void drawVideoRow(QPainter *painter, int index);
    void drawPlotRow(QPainter *painter);
    void drawVerticalTickMarks(QPainter *painter);
    void drawVerticalTickMark(QPainter *painter, qreal t, qreal value);
    void drawPlayhead(QPainter *painter);
    void drawScrollBar(QPainter *painter);
    void drawCurrentValueDots(QPainter *painter);

    void drawVerticalQuantityColorBar(QPainter *painter);

    // Math functions
    qreal timeToRectX(qreal time);
    qreal rectXToTime(qreal x);

    QPolygonF getPlayheadTriangle();

    static constexpr qreal timelineButtonSpacing = 5;
    static constexpr qreal leftLineOffsetAnalyzeMode = 50;
    static constexpr qreal playheadAreaHeight = 30;
    static constexpr qreal videoRowHeight = 25;
    static constexpr qreal scrollBarHeight = 15;
    static constexpr qreal scrollBarRadius = scrollBarHeight / 2.0;
    static constexpr qreal scrollBarHandlePadding = 2.5;

    bool m_holdingShift = false;
    qreal m_mousePressPoint = 0;
    QPointF m_previousMousePoint;

    enum class HoldingItem
    {
        None,
        Playhead,
        TimelineItem,
        ScrollBarLeftHandle,
        ScrollBarRightHandle,
        ScrollBarCenter
    };

    HoldingItem m_holding = HoldingItem::None;

    float videoToPlotRatio = 0.5f;

    float analyzeAnimationAmount = 0.0f;

    SVGColorRenderer m_lockFillRenderer;
    SVGColorRenderer m_lockDisabledRenderer;

    SVGColorRenderer m_visibilityFillRenderer;
    SVGColorRenderer m_visibilityDisabledRenderer;
    SVGColorRenderer m_earthRenderer;

    PlotDrawer* m_plotDrawer;
    QImage* m_plotImage = nullptr;
    QTimer* m_redrawPlotTimer;


public slots:
    void onPositionChanged();

    void setAnalyzeMode();
    void setSynchronizeMode();

    void onRedrawTimerExpired();

    void invalidatePlot();

protected slots:
    void onRedrawPlotsComplete(QImage* plotImage);
    void onContextMenuRequested(const QPoint &point);

signals:
    void selectionChanged();
};

#endif // TIMELINEVIEW_H
