#include "xyplotview.h"

XYPlotView::XYPlotView(QWidget *parent) : AbstractXYPlotView(parent)
{
    m_plotDrawer = new PlotDrawer(this, QSize(0,0), true);
    m_redrawPlotTimer = new QTimer(this);
    connect(m_redrawPlotTimer, &QTimer::timeout, this, &XYPlotView::onRedrawTimerExpired);
    m_redrawPlotTimer->setSingleShot(true);

    connect(m_plotDrawer, &PlotDrawer::plotsComplete, this, &XYPlotView::onRedrawPlotsComplete);

    connect(Engine::engine, &Engine::plotVisibilityChanged, this, &XYPlotView::onRedrawTimerExpired);
    connect(Engine::engine, &Engine::plotXQuantityIndexChanged, this, &XYPlotView::onRedrawTimerExpired);
    connect(Engine::engine, &Engine::plotYQuantityIndexChanged, [=]() { update(); });
    connect(Engine::engine, &Engine::positionChanged, [=]() { update(); });
    connect(Engine::engine, &Engine::contentsChanged, [=]() { update(); });
}

void XYPlotView::onRedrawPlotsComplete(QImage *plotImage)
{
    if (m_plotImage != nullptr) delete m_plotImage;
    m_plotImage = plotImage;
    update();
}

void XYPlotView::onRedrawTimerExpired()
{
    // Set up plot drawing thread
    m_plotDrawer->setSize(QSize(width(), height()));
    m_plotDrawer->start();
}

void XYPlotView::resizeEvent(QResizeEvent *event)
{
    QWidget::resizeEvent(event);
    invalidatePlot();
}

void XYPlotView::invalidatePlot()
{
    m_redrawPlotTimer->start(100);
}

QRect XYPlotView::getPlotRect()
{
    auto r = rect().marginsRemoved(QMargins(
                                     XY_BL_PADDING, // left
                                     XY_TR_PADDING, // top
                                     XY_TR_PADDING, // right
                                     XY_BL_PADDING  // bottom
                                     ));
    r.setWidth(qMax(0, r.width()));
    r.setHeight(qMax(0, r.height()));
    return r;
}

void XYPlotView::paintEvent(QPaintEvent *)
{
    QPainter painter(this);
    painter.setRenderHint(QPainter::RenderHint::Antialiasing, true);

    auto r = getPlotRect();

    if (m_plotImage != nullptr)
    {
        // Define rectangle where plot will be drawn
        painter.fillRect(r, QColorConstants::White);

        drawHorizontalTickMarks(&painter);
        drawVerticalTickMarks(&painter);

        painter.drawImage(r, *m_plotImage);
        drawCurrentValueDots(&painter);

        drawVerticalQuantityColorBar(&painter);
        drawHorizontalQuantityColorBar(&painter);

        // Draw border around XY plot
        painter.setBrush(Qt::NoBrush);
        painter.setPen(QColorConstants::Gray);
        painter.drawRect(r);
    }
    else
    {
        // Case of no image to draw - display message to the user instead
        QString message("No plot loaded");

        QFont font = painter.font();
        font.setItalic(true);
        painter.setFont(font);

        painter.setPen(DefaultColorConstants::TextSecondary);

        painter.drawText(rect(), Qt::AlignCenter, message);
    }

}

void XYPlotView::drawVerticalTickMarks(QPainter *painter)
{
    qreal cyPerInch = 5.;

    qreal physicalSize = qMax(0, getPlotRect().height());

    TickMarkData data = PlotMath::calculateTickMarks(
                Engine::engine->plotYQuantity(), true,
                cyPerInch, physicalSize);

    qreal tickLocation = data.startingTickLocation;
    qreal lineY = 0;
    while (lineY < 1)
    {
        lineY = invLerp(data.bottomBound, data.topBound, tickLocation);
        drawVerticalTickMark(painter, lineY, tickLocation);
        tickLocation += data.unitsPerGrid;
    }
}

void XYPlotView::drawHorizontalTickMarks(QPainter *painter)
{
    qreal cxPerInch = 10.;

    auto xQuant = Engine::engine->plotXQuantity();
    if (xQuant == nullptr)
        return;

    qreal physicalSize = qMax(0, getPlotRect().width());
    TickMarkData data = PlotMath::calculateTickMarks(
                xQuant, false,
                cxPerInch, physicalSize);

    qreal tickLocation = data.startingTickLocation;
    qreal lineX = 0;
    while (lineX < 1)
    {
        lineX = invLerp(data.bottomBound, data.topBound, tickLocation);
        drawHorizontalTickMark(painter, lineX, tickLocation);
        tickLocation += data.unitsPerGrid;
    }
}

void XYPlotView::drawHorizontalTickMark(QPainter *painter, qreal t, qreal value)
{
    auto r = getPlotRect();
    qint32 leftXWidgetSpace = lerp(r.x(), r.x() + r.width(), t);
    QPoint p1(leftXWidgetSpace, r.y() + r.height());
    QPoint p2(leftXWidgetSpace, r.bottom() + TICK_LENGTH);
    QPoint p3(leftXWidgetSpace, r.top());

    painter->setPen(DefaultColorConstants::LineSecondary);
    painter->drawLine(p2, p3);

    painter->setPen(QColorConstants::Gray);
    painter->drawLine(p1, p2);

    painter->setPen(QColorConstants::Black);

    // Compute position to draw label
    QString labelText = QString::number(value);
    QRect textRect = painter->fontMetrics().boundingRect(labelText);
    painter->drawText(p2.x() - textRect.width() / 2, p2.y() + textRect.height(), labelText);
}

void XYPlotView::drawVerticalTickMark(QPainter *painter, qreal t, qreal value)
{
    auto r = getPlotRect();
    qint32 leftYWidgetSpace = lerp(r.y() + r.height(), r.y(), t);
    QPoint p1(getPlotRect().left(), leftYWidgetSpace);
    QPoint p2(getPlotRect().left() - TICK_LENGTH, leftYWidgetSpace);
    QPoint p3(getPlotRect().right(), leftYWidgetSpace);

    painter->setPen(DefaultColorConstants::LineSecondary);
    painter->drawLine(p2, p3);

    painter->setPen(QColorConstants::Gray);
    painter->drawLine(p1, p2);

    painter->setPen(QColorConstants::Black);

    // Compute position to draw label
    QString labelText = QString::number(value);
    QRect textRect = painter->fontMetrics().boundingRect(labelText);
    painter->drawText(p2.x() - textRect.width() - TICK_PADDING, p2.y() + textRect.height() / 2, labelText);
}

void XYPlotView::drawCurrentValueDots(QPainter *painter)
{
    qreal plotTime = Engine::engine->currentTime() - Engine::engine->plotDisplayOffset();

    auto plot = Engine::engine->currentPlot();

    if (plot == nullptr)
    {
        qDebug() << "drawCurrentValueDots() - plot is nullptr";
        return;
    }

    auto r = getPlotRect();
    for (auto& quantity : plot->quantities())
    {
        qreal xPos = plot->relativeValueForQuantityAtTime(Engine::engine->plotXQuantity(), plotTime);
        qreal yPos = plot->relativeValueForQuantityAtTime(quantity, plotTime);

        qreal top = r.y();
        qreal bottom = r.y() + r.height();
        if (!plot->scaleGroups().value(quantity->scaleGroup())->railToRail())
        {
            qreal padding = (top - bottom) * (1 - RAIL_TO_RAIL_RATIO);
            top -= padding;
            bottom += padding;
        }

        qreal left = r.x();
        qreal right = r.x() + r.width();

        painter->setBrush(quantity->color());
        painter->setPen(QPen(QColor(quantity->color()).darker(140), 1));

        qreal dotRadius = DOT_RADIUS;
        painter->drawEllipse(lerp(left, right, xPos) - dotRadius,
                             lerp(bottom, top, yPos) - dotRadius,
                             dotRadius * 2, dotRadius * 2);
    }
}

void XYPlotView::drawVerticalQuantityColorBar(QPainter *painter)
{
    auto r = getPlotRect();
    qreal barWidth = QUANTITY_BAR_THICKNESS;
    auto yAxisQuantities = Engine::engine->quantitiesForCurrentScaleGroup();
    auto singleRectHeight = (getPlotRect().height() - 2 * barWidth) / yAxisQuantities.length();
    QRectF singleColorRect = QRect(0, r.y(), barWidth, singleRectHeight);
    painter->setPen(Qt::NoPen);
    for (int i = 0; i < yAxisQuantities.length(); i++)
    {
        auto thisRect = singleColorRect.translated(0, barWidth + singleRectHeight * i);

        // Make color
        auto quantityColor = yAxisQuantities.at(i)->color();
        painter->setBrush(quantityColor);
        painter->drawRect(thisRect);
        auto r = thisRect.width() / 2;

        // First rectangle should have a round cap on top
        if (i == 0)
        {
            auto topCenter = QPointF(thisRect.center().x(), thisRect.y());
            painter->drawEllipse(topCenter, r, r);
        }

        // Last rectangle should have a round cap on bottom
        // (Note: this can't be an else if statement, because it's possible
        // that there will be only one rectangle, and thus that there should
        // be both caps.)
        if (i == yAxisQuantities.length() - 1)
        {
            auto bottomCenter = QPointF(thisRect.center().x(), thisRect.y() + thisRect.height());
            painter->drawEllipse(bottomCenter, r, r);
        }
    }
}

void XYPlotView::drawHorizontalQuantityColorBar(QPainter *painter)
{
    qreal barHeight = QUANTITY_BAR_THICKNESS;
    auto xAxisQuantities = Engine::engine->quantitiesForCurrentXAxisScaleGroup();
    auto singleRectWidth = (getPlotRect().width() - 2 * barHeight) / xAxisQuantities.length();
    QRectF singleColorRect = QRect(getPlotRect().x(), getPlotRect().y() + getPlotRect().height() + 25, singleRectWidth, barHeight);
    painter->setPen(Qt::NoPen);
    for (int i = 0; i < xAxisQuantities.length(); i++)
    {
        auto thisRect = singleColorRect.translated(barHeight + singleRectWidth * i, 0);

        // Make color
        auto quantityColor = xAxisQuantities.at(i)->color();
        painter->setBrush(quantityColor);
        painter->drawRect(thisRect);
        auto r = thisRect.height() / 2;

        // First rectangle should have a round cap on top
        if (i == 0)
        {
            auto topCenter = QPointF(thisRect.x(), thisRect.center().y());
            painter->drawEllipse(topCenter, r, r);
        }

        // Last rectangle should have a round cap on bottom
        // (Note: this can't be an else if statement, because it's possible
        // that there will be only one rectangle, and thus that there should
        // be both caps.)
        if (i == xAxisQuantities.length() - 1)
        {
            auto bottomCenter = QPointF(thisRect.x() + thisRect.width(), thisRect.center().y());
            painter->drawEllipse(bottomCenter, r, r);
        }
    }
}
